import errno
import os
import pathlib
import random
import time
from unittest.mock import call, patch

import pytest

from upsies import __project_name__, errors
from upsies.utils import fs


def test_assert_file_readable_with_directory(tmp_path):
    with pytest.raises(errors.ContentError, match=rf'^{tmp_path}: Is a directory$'):
        fs.assert_file_readable(tmp_path)

def test_assert_file_readable_with_nonexisting_file(tmp_path):
    path = tmp_path / 'foo'
    with pytest.raises(errors.ContentError, match=rf'^{path}: No such file or directory$'):
        fs.assert_file_readable(path)

def test_assert_file_readable_with_unreadable_file(tmp_path):
    path = tmp_path / 'foo'
    path.write_text('bar')
    os.chmod(path, mode=0o000)
    try:
        with pytest.raises(errors.ContentError, match=rf'^{path}: Permission denied$'):
            fs.assert_file_readable(path)
    finally:
        os.chmod(path, mode=0o700)


def test_assert_dir_usable_with_file(tmp_path):
    path = tmp_path / 'foo'
    path.write_text('asdf')
    with pytest.raises(errors.ContentError, match=rf'^{path}: Not a directory$'):
        fs.assert_dir_usable(path)

def test_assert_dir_usable_with_nonexisting_directory(tmp_path):
    path = tmp_path / 'foo'
    with pytest.raises(errors.ContentError, match=rf'^{path}: Not a directory$'):
        fs.assert_dir_usable(path)

def test_assert_dir_usable_with_unreadable_directory(tmp_path):
    path = tmp_path / 'foo'
    path.mkdir()
    os.chmod(path, mode=0o333)
    try:
        with pytest.raises(errors.ContentError, match=rf'^{path}: Not readable$'):
            fs.assert_dir_usable(path)
    finally:
        os.chmod(path, mode=0o700)

def test_assert_dir_usable_with_unwritable_directory(tmp_path):
    path = tmp_path / 'foo'
    path.mkdir()
    os.chmod(path, mode=0o555)
    try:
        with pytest.raises(errors.ContentError, match=rf'^{path}: Not writable$'):
            fs.assert_dir_usable(path)
    finally:
        os.chmod(path, mode=0o700)

def test_assert_dir_usable_with_unexecutable_directory(tmp_path):
    path = tmp_path / 'foo'
    path.mkdir()
    os.chmod(path, mode=0o666)
    try:
        with pytest.raises(errors.ContentError, match=rf'^{path}: Not executable$'):
            fs.assert_dir_usable(path)
    finally:
        os.chmod(path, mode=0o700)

def test_assert_dir_usable_with_usable_directory(tmp_path):
    return_value = fs.assert_dir_usable(tmp_path)
    assert return_value is None


@pytest.mark.parametrize(
    argnames='content_path, base, exp_projectdir',
    argvalues=(
        ('path/to/foo', None, f'default/cache/directory/foo.{__project_name__}'),
        ('path/to//foo', None, f'default/cache/directory/foo.{__project_name__}'),
        ('path/to/foo/', None, f'default/cache/directory/foo.{__project_name__}'),
        ('path/to/foo', 'my/path', f'my/path/foo.{__project_name__}'),
        (None, None, '.'),
        (None, 'my/path', '.'),
        ('', None, '.'),
        ('', 'my/path', '.'),
    ),
)
def test_projectdir(content_path, base, exp_projectdir, mocker, tmp_path):
    mocker.patch('upsies.constants.DEFAULT_CACHE_DIRECTORY', 'default/cache/directory')
    mkdir_mock = mocker.patch('upsies.utils.fs.mkdir')

    fs.projectdir.cache_clear()
    path = fs.projectdir(content_path, base=base)
    assert path == exp_projectdir

    if content_path:
        assert mkdir_mock.call_args_list == [call(path)]
    else:
        assert mkdir_mock.call_args_list == []


def test_limit_directory_size(tmp_path):
    for dirname1 in ('a', 'b', 'c'):
        for dirname2 in ('d', 'e', 'f'):
            (tmp_path / dirname1 / dirname2).mkdir(parents=True)
            (tmp_path / dirname1 / dirname2 / 'y').write_text('_' * 100)
            (tmp_path / dirname1 / dirname2 / 'empty_file').write_text('')
            (tmp_path / dirname1 / dirname2 / 'empty_dir').mkdir()
            (tmp_path / dirname1 / dirname2 / 'symlink_to_empty_file').symlink_to('empty_file')
            (tmp_path / dirname1 / dirname2 / 'symlink_to_nonempty_file').symlink_to('y')
            (tmp_path / dirname1 / dirname2 / 'symlink_to_empty_dir').symlink_to('empty_dir')
            (tmp_path / dirname1 / dirname2 / 'symlink_to_nonempty_dir').symlink_to('..')
            (tmp_path / dirname1 / dirname2 / 'dead_symlink').symlink_to('the void')
        (tmp_path / dirname1 / 'z').write_text('_' * 1000)
        (tmp_path / dirname1 / 'empty_file').write_text('')
        (tmp_path / dirname1 / 'empty_dir').mkdir()

    now = time.time()
    min_age = 125
    max_age = 300
    atime = now - max_age
    age_diff = 5
    for dirname1 in ('a', 'b', 'c'):
        for dirname2 in ('d', 'e', 'f'):
            for entry in ('y', 'empty_file', 'empty_dir'):
                atime += age_diff
                os.utime(tmp_path / dirname1 / dirname2 / entry, (atime, atime))
        for entry in ('z', 'empty_file', 'empty_dir'):
            atime += age_diff
            os.utime(tmp_path / dirname1 / entry, (atime, atime))

    (tmp_path / 'a' / 'd' / 'too_old').write_text('x')
    os.utime(tmp_path / 'a' / 'd' / 'too_old', (now - max_age - 1, now - max_age - 1))
    (tmp_path / 'c' / 'too_young').write_text('y')
    os.utime(tmp_path / 'c' / 'too_young', (now - min_age + 1, now - min_age + 1))

    def get_stat(path):
        return os.stat(path, follow_symlinks=False)

    def get_size(path):
        if os.path.islink(path):
            return 0
        try:
            return os.path.getsize(path)
        except OSError:
            return 0

    def get_total_size():
        return sum(get_size(f) for f in get_files())

    def get_files():
        return sorted(
            (
                os.path.join(dirpath, filename)
                for dirpath, dirnames, filenames in os.walk(tmp_path)
                for filename in filenames
                if get_size(os.path.join(dirpath, filename))
            ),
            key=lambda filepath: get_stat(filepath).st_atime,
        )

    def print_current_files():
        for f in sorted(get_files(), key=lambda f: get_stat(f).st_atime):
            print(f'{f:<72}', 'age:', round(now - get_stat(f).st_atime), 'size:', get_size(f))
        print(' ' * 75, 'total size:', get_total_size())

    orig_files = get_files()
    print('initial files:')
    print_current_files()

    # No pruning necessary
    fs.limit_directory_size(tmp_path, max_total_size=3900, min_age=min_age, max_age=max_age)
    print('after max_total_size=3900:')
    print_current_files()
    assert get_files() == orig_files
    assert get_total_size() == 3902

    # Prune oldest file
    fs.limit_directory_size(tmp_path, max_total_size=3800, min_age=min_age, max_age=max_age)
    print('after max_total_size=3800:')
    print_current_files()
    assert get_files() == (
        [orig_files[0]]  # too_old
        + orig_files[2:]
    )
    assert get_total_size() == 3802

    # Prune multiple files
    fs.limit_directory_size(tmp_path, max_total_size=3000, min_age=min_age, max_age=max_age)
    print('after max_total_size=3000:')
    print_current_files()
    assert get_files() == (
        [orig_files[0]]  # too_old
        + orig_files[5:]
    )
    assert get_total_size() == 2602

    # Prune all files
    fs.limit_directory_size(tmp_path, max_total_size=0, min_age=min_age, max_age=max_age)
    assert get_files() == (
        [orig_files[0]]     # too_old
        + [orig_files[-1]]  # too_young
    )
    assert get_total_size() == 2

def test_limit_directory_size_with_nonexisting_path(tmp_path):
    fs.limit_directory_size(tmp_path / 'does' / 'not' / 'exist', max_total_size=123)
    assert not os.path.exists(tmp_path / 'does' / 'not' / 'exist')


def test_prune_empty_prunes_empty_files(tmp_path):
    (tmp_path / 'bar' / 'x').mkdir(parents=True)
    (tmp_path / 'bar' / 'y' / 'z' / '1').mkdir(parents=True)
    (tmp_path / 'bar' / 'y' / 'z' / '1' / 'c').write_text('yes, this is c')
    (tmp_path / 'bar' / 'y' / 'z' / '1' / 'also_empty').write_text('')
    (tmp_path / 'bar' / 'y' / 'z' / '2').mkdir(parents=True)
    (tmp_path / 'bar' / 'y' / 'z' / 'symlink_to_empty_directory').symlink_to('../../../foo/a/1')
    (tmp_path / 'bar' / 'y' / 'z' / 'symlink_to_nonempty_directory').symlink_to('../../../foo/')
    (tmp_path / 'bar' / 'y' / 'z' / 'symlink_to_empty_file').symlink_to('../../../foo/empty')
    (tmp_path / 'bar' / 'y' / 'z' / 'symlink_to_nonempty_file').symlink_to('../../../foo/b')
    (tmp_path / 'baz').mkdir(parents=True)
    (tmp_path / 'foo' / 'a' / '1').mkdir(parents=True)
    (tmp_path / 'foo' / 'a' / '2').mkdir(parents=True)
    (tmp_path / 'foo' / 'b').write_text('yes, this is b')
    (tmp_path / 'foo' / 'empty').write_text('')
    (tmp_path / 'dead_symlink').symlink_to('nothing')
    fs.prune_empty(tmp_path, files=True, directories=False)
    assert os.path.exists(tmp_path)
    assert os.path.isdir(tmp_path)
    tree = []
    for dirpath, dirnames, filenames in os.walk(tmp_path):
        for dirname in dirnames:
            tree.append(os.path.join(dirpath, dirname))
        for filename in filenames:
            tree.append(os.path.join(dirpath, filename))
    assert sorted(tree) == [
        f'{tmp_path}/bar',
        f'{tmp_path}/bar/x',
        f'{tmp_path}/bar/y',
        f'{tmp_path}/bar/y/z',
        f'{tmp_path}/bar/y/z/1',
        f'{tmp_path}/bar/y/z/1/c',
        f'{tmp_path}/bar/y/z/2',
        f'{tmp_path}/bar/y/z/symlink_to_empty_directory',
        f'{tmp_path}/bar/y/z/symlink_to_nonempty_directory',
        f'{tmp_path}/bar/y/z/symlink_to_nonempty_file',
        f'{tmp_path}/baz',
        f'{tmp_path}/foo',
        f'{tmp_path}/foo/a',
        f'{tmp_path}/foo/a/1',
        f'{tmp_path}/foo/a/2',
        f'{tmp_path}/foo/b',
    ]

def test_prune_empty_prunes_empty_directories(tmp_path):
    (tmp_path / 'bar' / 'x').mkdir(parents=True)
    (tmp_path / 'bar' / 'y' / 'z' / '1').mkdir(parents=True)
    (tmp_path / 'bar' / 'y' / 'z' / '1' / 'c').write_text('yes, this is c')
    (tmp_path / 'bar' / 'y' / 'z' / '1' / 'empty').write_text('')
    (tmp_path / 'bar' / 'y' / 'z' / '2').mkdir(parents=True)
    (tmp_path / 'bar' / 'y' / 'z' / 'symlink_to_empty_directory').symlink_to('../../../foo/a/1')
    (tmp_path / 'bar' / 'y' / 'z' / 'symlink_to_nonempty_directory').symlink_to('../../../foo/')
    (tmp_path / 'bar' / 'y' / 'z' / 'symlink_to_empty_file').symlink_to('1/empty')
    (tmp_path / 'bar' / 'y' / 'z' / 'symlink_to_nonempty_file').symlink_to('../../../foo/b')
    (tmp_path / 'baz').mkdir(parents=True)
    (tmp_path / 'foo' / 'a' / '1').mkdir(parents=True)
    (tmp_path / 'foo' / 'a' / '2').mkdir(parents=True)
    (tmp_path / 'foo' / 'b').write_text('yes, this is b')
    (tmp_path / 'foo' / 'also_empty').write_text('')
    (tmp_path / 'dead_symlink').symlink_to('nothing')
    fs.prune_empty(tmp_path, files=False, directories=True)
    assert os.path.exists(tmp_path)
    assert os.path.isdir(tmp_path)
    tree = []
    for dirpath, dirnames, filenames in os.walk(tmp_path):
        for dirname in dirnames:
            tree.append(os.path.join(dirpath, dirname))
        for filename in filenames:
            tree.append(os.path.join(dirpath, filename))
    assert sorted(tree) == [
        f'{tmp_path}/bar',
        f'{tmp_path}/bar/y',
        f'{tmp_path}/bar/y/z',
        f'{tmp_path}/bar/y/z/1',
        f'{tmp_path}/bar/y/z/1/c',
        f'{tmp_path}/bar/y/z/1/empty',
        f'{tmp_path}/bar/y/z/symlink_to_empty_file',
        f'{tmp_path}/bar/y/z/symlink_to_nonempty_directory',
        f'{tmp_path}/bar/y/z/symlink_to_nonempty_file',
        f'{tmp_path}/foo',
        f'{tmp_path}/foo/also_empty',
        f'{tmp_path}/foo/b',
    ]

def test_prune_empty_prunes_root_directory(tmp_path):
    (tmp_path / 'foo' / 'a' / '1').mkdir(parents=True)
    (tmp_path / 'foo' / 'a' / '2').mkdir(parents=True)
    (tmp_path / 'bar' / 'x').mkdir(parents=True)
    (tmp_path / 'bar' / 'y' / 'z' / '1').mkdir(parents=True)
    (tmp_path / 'bar' / 'y' / 'z' / '2').mkdir(parents=True)
    (tmp_path / 'baz').mkdir(parents=True)
    fs.prune_empty(tmp_path)
    assert not os.path.exists(tmp_path)

@pytest.mark.parametrize('exc_args', ((os.strerror(errno.EACCES),), (errno.EACCES, os.strerror(errno.EACCES),)))
@pytest.mark.parametrize(
    argnames='error_function, kwargs, exp_subpath',
    argvalues=(
        ('upsies.utils.fs.file_size', {'files': True}, 'empty_file'),
        ('os.unlink', {'files': True}, 'empty_file'),
        ('os.listdir', {'directories': True}, 'empty_dir'),
        ('os.rmdir', {'directories': True}, 'empty_dir'),
    ),
    ids=lambda v: str(v),
)
def test_prune_empty_handles_OSError(error_function, kwargs, exp_subpath, exc_args, tmp_path, mocker):
    mocker.patch(error_function, side_effect=OSError(*exc_args))
    (tmp_path / 'empty_file').write_text('')
    (tmp_path / 'empty_dir').mkdir()
    with pytest.raises(RuntimeError, match=rf'{tmp_path}/{exp_subpath}: Failed to prune: Permission denied'):
        fs.prune_empty(tmp_path, **kwargs)

@pytest.mark.parametrize('exc_args', ((os.strerror(errno.EACCES),), (errno.EACCES, os.strerror(errno.EACCES),)))
def test_prune_empty_handles_OSError_from_deleting_root_directory(exc_args, tmp_path, mocker):
    mocker.patch('os.rmdir', side_effect=OSError(*exc_args))
    with pytest.raises(RuntimeError, match=rf'{tmp_path}: Failed to prune: Permission denied'):
        fs.prune_empty(tmp_path)

def test_prune_empty_handles_nonexisting_path(tmp_path):
    fs.prune_empty(tmp_path / 'does' / 'not' / 'exist')
    assert not os.path.exists(tmp_path / 'does' / 'not' / 'exist')

def test_prune_empty_handles_file_path(tmp_path):
    (tmp_path / 'foo').write_text('not a directory')
    fs.prune_empty(tmp_path / 'foo')
    assert os.path.exists(tmp_path / 'foo')


@patch('os.makedirs')
@patch('upsies.utils.fs.assert_dir_usable')
def test_mkdir(assert_dir_usable_mock, makedirs_mock):
    fs.mkdir('path/to/dir')
    assert makedirs_mock.call_args_list == [call('path/to/dir', exist_ok=True)]
    assert assert_dir_usable_mock.call_args_list == [call('path/to/dir')]

@patch('os.makedirs')
@patch('upsies.utils.fs.assert_dir_usable')
def test_mkdir_handles_empty_path(assert_dir_usable_mock, makedirs_mock):
    fs.mkdir('')
    assert makedirs_mock.call_args_list == []
    assert assert_dir_usable_mock.call_args_list == []

@patch('os.makedirs')
@patch('upsies.utils.fs.assert_dir_usable')
def test_mkdir_catches_makedirs_error(assert_dir_usable_mock, makedirs_mock):
    makedirs_mock.side_effect = OSError('No way')
    with pytest.raises(errors.ContentError, match=r'^path/to/dir: No way'):
        fs.mkdir('path/to/dir')
    assert makedirs_mock.call_args_list == [call('path/to/dir', exist_ok=True)]
    assert assert_dir_usable_mock.call_args_list == []


@pytest.mark.parametrize(
    argnames='path, parents, exp_basename',
    argvalues=(
        ('a/b/c', 0, 'c'),
        ('a/b/c', 1, 'b/c'),
        ('a/b/c', 2, 'a/b/c'),
        ('a/b/c', 3, 'a/b/c'),

        ('a/b/c/', 0, 'c'),
        ('a/b/c/', 1, 'b/c'),
        ('a/b/c/', 2, 'a/b/c'),
        ('a/b/c/', 3, 'a/b/c'),

        ('a//b/c///', 0, 'c'),
        ('a//b/c///', 1, 'b/c'),
        ('a//b/c///', 2, 'a/b/c'),
        ('a//b/c///', 3, 'a/b/c'),

        (pathlib.Path('a/b/c//d/'), 0, 'd'),
        (pathlib.Path('a/b/c//d/'), 1, 'c/d'),
        (pathlib.Path('a/b/c//d/'), 2, 'b/c/d'),
        (pathlib.Path('a/b/c//d/'), 3, 'a/b/c/d'),
        (pathlib.Path('a/b/c//d/'), 4, 'a/b/c/d'),
    ),
    ids=lambda v: repr(v),
)
def test_basename(path, parents, exp_basename):
    assert fs.basename(path, parents=parents) == exp_basename


def test_dirname():
    import pathlib
    assert fs.dirname('a/b/c') == 'a/b'
    assert fs.dirname('a/b/c/') == 'a/b'
    assert fs.dirname('a/b/c///') == 'a/b'
    assert fs.dirname('a/b/c//d/') == 'a/b/c'
    assert fs.dirname(pathlib.Path('a/b/c//d/')) == 'a/b/c'


def test_file_and_parent():
    assert fs.file_and_parent('a/b/c//d') == ('d', 'c')
    assert fs.file_and_parent('a/b/c//d/') == ('d', 'c')
    assert fs.file_and_parent('d') == ('d',)


def test_sanitize_filename_on_unix(mocker):
    mocker.patch('upsies.utils.fs.os_family', return_value='unix')
    assert fs.sanitize_filename('foo<bar>baz :a"b/c \\1|2?3*4.txt') == 'foo<bar>baz :a"b_c \\1|2?3*4.txt'

def test_sanitize_filename_on_windows(mocker):
    mocker.patch('upsies.utils.fs.os_family', return_value='windows')
    assert fs.sanitize_filename('foo<bar>baz :a"b/c \\1|2?3*4.txt') == 'foo_bar_baz _a_b_c _1_2_3_4.txt'


@pytest.mark.parametrize(
    argnames='os_family, sep, path, exp_sanitized_path',
    argvalues=(
        ('unix', '/', 'C:\\foo<bar>baz :a"b/c \\ 1|2?3*4.txt', 'C:\\foo<bar>baz :a"b/c \\ 1|2?3*4.txt'),
        ('unix', '/', '/foo<bar>baz :a"b/c \\ 1|2?3*4.txt', '/foo<bar>baz :a"b/c \\ 1|2?3*4.txt'),
        ('windows', '\\', 'C:\\foo<bar>baz :a"b/c \\ 1|2?3*4.txt', 'C:\\foo_bar_baz _a_b_c \\ 1_2_3_4.txt'),
        ('windows', '\\', '/foo<bar>baz :a"b/c \\ 1|2?3*4.txt', '/foo<bar>baz :a_b_c \\ 1_2_3_4.txt'),
    ),
)
def test_sanitize_path(os_family, sep, path, exp_sanitized_path, mocker):
    mocker.patch('upsies.utils.fs.os_family', return_value=os_family)
    if os_family == 'unix':
        mocker.patch('os.path.splitdrive', return_value=('', path))
    elif os_family == 'windows':
        drive_letter, path = path.split(':', maxsplit=1)
        mocker.patch('os.path.splitdrive', return_value=(f'{drive_letter}:', path))
    else:
        mocker.patch('os.path.splitdrive', side_effect=RuntimeError('Trying to add support for a new OS?'))

    import os
    mocker.patch.object(os, 'sep', sep)
    sanitized_path = fs.sanitize_path(path)
    assert sanitized_path == exp_sanitized_path


@pytest.mark.parametrize(
    argnames='home, path, exp_path',
    argvalues=(
        ('/home/foo', 'relative/path', 'relative/path'),
        ('/home/foo', '/absolute/path', '/absolute/path'),
        ('/home/foo', '/home/foo/path', '~/path'),
        ('/home/foo', '/absolute/path/to/home/foo/bar', '/absolute/path/to/home/foo/bar'),
        ('/home/foo(with]regex}chars', '/home/foo(with]regex}chars/bar/baz', '~/bar/baz'),
    ),
)
def test_tildify_path(home, path, exp_path, mocker):
    mocker.patch('os.path.expanduser', return_value=home)
    assert fs.tildify_path(path) == exp_path


@pytest.mark.parametrize(
    argnames='path, minlen, maxlen, only, exp_name, exp_extension',
    argvalues=(
        ('path/to/Something.x264-GRP.mkv', None, None, [], 'path/to/Something.x264-GRP', 'mkv'),
        ('path/to/Something.x264-GRP.MKV', None, None, [], 'path/to/Something.x264-GRP', 'mkv'),
        ('path/to/Something.x264-GRP.mp4', None, None, [], 'path/to/Something.x264-GRP', 'mp4'),
        ('path/to/Something.x264-GRP.Mp4', None, None, ['mKv'], 'path/to/Something.x264-GRP.Mp4', ''),
        ('path/to/Something.x264-GRP.Mkv', None, None, ['mkV'], 'path/to/Something.x264-GRP', 'mkv'),
        ('path/to/Something.x264-GRP', None, None, [], 'path/to/Something.x264-GRP', ''),
        (pathlib.Path('some/path') / 'to' / 'file.mkv', None, None, [], 'some/path/to/file', 'mkv'),
        ('path/to/Something.x264-GRP.torrent', None, None, [], 'path/to/Something.x264-GRP', 'torrent'),
        ('path/to/Something.x264-GRP.torrent', None, 7, [], 'path/to/Something.x264-GRP', 'torrent'),
        ('path/to/Something.x264-GRP.torrent', None, 8, [], 'path/to/Something.x264-GRP', 'torrent'),
        ('path/to/Something.x264-GRP.torrent', None, 6, [], 'path/to/Something.x264-GRP.torrent', ''),
        ('path/to/Something.x264-GRP.x', 1, None, [], 'path/to/Something.x264-GRP', 'x'),
        ('path/to/Something.x264-GRP.x', 2, None, [], 'path/to/Something.x264-GRP.x', ''),
        ('path/to/Something.x264-GRP.jpg', 3, 3, [], 'path/to/Something.x264-GRP', 'jpg'),
        ('path/to/Something.x264-GRP.jpeg', 3, 3, [], 'path/to/Something.x264-GRP.jpeg', ''),
        ('path/to/Something.x264-GRP.jpeg', 3, 4, [], 'path/to/Something.x264-GRP', 'jpeg'),
        ('path/to/Something.x264-GRP.jpeg', 3, 4, [], 'path/to/Something.x264-GRP', 'jpeg'),
    ),
    ids=lambda v: repr(v),
)
def test_file_extension_and_strip_extension(path, minlen, maxlen, only, exp_name, exp_extension):
    assert fs.strip_extension(path, minlen=minlen, maxlen=maxlen, only=only) == exp_name
    assert fs.file_extension(path, minlen=minlen, maxlen=maxlen, only=only) == exp_extension


def test_file_size_of_file(tmp_path):
    filepath = tmp_path / 'file'
    filepath.write_bytes(b'foo')
    assert fs.file_size(filepath) == 3

def test_file_size_of_directory(tmp_path):
    assert fs.file_size(tmp_path) is None

def test_file_size_of_unaccessible_file(tmp_path):
    filepath = tmp_path / 'file'
    filepath.write_bytes(b'foo')
    os.chmod(tmp_path, 0o000)
    try:
        assert fs.file_size(filepath) is None
    finally:
        os.chmod(tmp_path, 0o700)

def test_file_size_of_nonexisting_file():
    assert fs.file_size('path/to/nothing') is None


def test_path_size(tmp_path):
    tree = [
        ('a', 0),
        ('b', 1),
        ('c/1', 2),
        ('c/2', 3),
        ('c/d/x', 4),
        ('c/d/y', 5),
        ('c/d/z', 6),
        ('l/1', 'c/1'),
        ('l/2', 'c/d/y'),
    ]
    random.shuffle(tree)
    for f, size_or_linktarget in tree:
        fpath = tmp_path / f
        fpath.parent.mkdir(parents=True, exist_ok=True)
        if isinstance(size_or_linktarget, int):
            print('writing', size_or_linktarget, 'bytes to', fpath)
            fpath.write_bytes(b'x' * size_or_linktarget)
        else:
            link_target = tmp_path / size_or_linktarget
            print('linking', fpath, 'to', link_target)
            fpath.symlink_to(link_target)

    size = fs.path_size(tmp_path)
    assert size == 21


def test_file_list_recurses_into_subdirectories(tmp_path):
    (tmp_path / 'a.txt').write_bytes(b'foo')
    (tmp_path / 'a').mkdir()
    (tmp_path / 'a' / 'b.txt').write_bytes(b'foo')
    (tmp_path / 'a' / 'b').mkdir()
    (tmp_path / 'a' / 'b' / 'b.txt').write_bytes(b'foo')
    (tmp_path / 'a' / 'b' / 'c').mkdir()
    (tmp_path / 'a' / 'b' / 'c' / 'c.txt').write_bytes(b'foo')
    (tmp_path / 'a' / 'b' / 'dead_symlink').symlink_to('not that')
    (tmp_path / 'a' / 'b' / 'symlink').symlink_to('c/c.txt')
    assert fs.file_list(tmp_path) == (
        str((tmp_path / 'a.txt')),
        str((tmp_path / 'a' / 'b.txt')),
        str((tmp_path / 'a' / 'b' / 'b.txt')),
        str((tmp_path / 'a' / 'b' / 'c' / 'c.txt')),
        str((tmp_path / 'a' / 'b' / 'dead_symlink')),
        str((tmp_path / 'a' / 'b' / 'symlink')),
    )

def test_file_list_with_follow_dirlinks(tmp_path):
    (tmp_path / 'directory').mkdir()
    (tmp_path / 'directory' / 'directory_contents').write_bytes(b'foo')
    (tmp_path / 'link_to_directory').symlink_to('directory')
    (tmp_path / 'file').write_bytes(b'foo')
    (tmp_path / 'link_to_file').symlink_to('file')
    (tmp_path / 'dead_link').symlink_to('no_such_file')
    assert fs.file_list(tmp_path, follow_dirlinks=True) == (
        str((tmp_path / 'dead_link')),
        str((tmp_path / 'directory' / 'directory_contents')),
        str((tmp_path / 'file')),
        str((tmp_path / 'link_to_directory' / 'directory_contents')),
        str((tmp_path / 'link_to_file')),
    )

def test_file_list_without_follow_dirlinks(tmp_path):
    (tmp_path / 'directory').mkdir()
    (tmp_path / 'directory' / 'directory_contents').write_bytes(b'foo')
    (tmp_path / 'link_to_directory').symlink_to('directory')
    (tmp_path / 'file').write_bytes(b'foo')
    (tmp_path / 'link_to_file').symlink_to('file')
    (tmp_path / 'dead_link').symlink_to('no_such_file')
    assert fs.file_list(tmp_path, follow_dirlinks=False) == (
        str((tmp_path / 'dead_link')),
        str((tmp_path / 'directory' / 'directory_contents')),
        str((tmp_path / 'file')),
        str((tmp_path / 'link_to_directory')),
        str((tmp_path / 'link_to_file')),
    )

def test_file_list_sorts_files_naturally(tmp_path):
    (tmp_path / '3' / '500').mkdir(parents=True)
    (tmp_path / '3' / '500' / '1000.txt').write_bytes(b'foo')
    (tmp_path / '20').mkdir()
    (tmp_path / '20' / '9.txt').write_bytes(b'foo')
    (tmp_path / '20' / '10.txt').write_bytes(b'foo')
    (tmp_path / '20' / '99').mkdir()
    (tmp_path / '20' / '99' / '4.txt').write_bytes(b'foo')
    (tmp_path / '20' / '99' / '0001000.txt').write_bytes(b'foo')
    (tmp_path / '20' / '100').mkdir()
    (tmp_path / '20' / '100' / '7.txt').write_bytes(b'foo')
    (tmp_path / '20' / '100' / '300.txt').write_bytes(b'foo')
    assert fs.file_list(tmp_path) == (
        str((tmp_path / '3' / '500' / '1000.txt')),
        str((tmp_path / '20' / '9.txt')),
        str((tmp_path / '20' / '10.txt')),
        str((tmp_path / '20' / '99' / '4.txt')),
        str((tmp_path / '20' / '99' / '0001000.txt')),
        str((tmp_path / '20' / '100' / '7.txt')),
        str((tmp_path / '20' / '100' / '300.txt')),
    )

def test_file_list_filters_by_extensions(tmp_path):
    (tmp_path / 'bar.jpg').write_bytes(b'foo')
    (tmp_path / 'foo.txt').write_bytes(b'foo')
    (tmp_path / 'a').mkdir()
    (tmp_path / 'a' / 'a1.txt').write_bytes(b'foo')
    (tmp_path / 'a' / 'a2.jpg').write_bytes(b'foo')
    (tmp_path / 'a' / 'b').mkdir()
    (tmp_path / 'a' / 'b' / 'b1.JPG').write_bytes(b'foo')
    (tmp_path / 'a' / 'b' / 'b2.TXT').write_bytes(b'foo')
    (tmp_path / 'a' / 'b' / 'c').mkdir()
    (tmp_path / 'a' / 'b' / 'c' / 'c1.Txt').write_bytes(b'foo')
    (tmp_path / 'a' / 'b' / 'c' / 'c2.jPG').write_bytes(b'foo')
    (tmp_path / 'a' / 'b' / 'c' / 'd').mkdir()
    (tmp_path / 'a' / 'b' / 'c' / 'd' / 'd1.txt').write_bytes(b'foo')
    (tmp_path / 'a' / 'b' / 'c' / 'd' / 'd2.txt').write_bytes(b'foo')
    assert fs.file_list(tmp_path, extensions=('jpg',)) == (
        str((tmp_path / 'a' / 'a2.jpg')),
        str((tmp_path / 'a' / 'b' / 'b1.JPG')),
        str((tmp_path / 'a' / 'b' / 'c' / 'c2.jPG')),
        str((tmp_path / 'bar.jpg')),
    )
    assert fs.file_list(tmp_path, extensions=('TXT',)) == (
        str((tmp_path / 'a' / 'a1.txt')),
        str((tmp_path / 'a' / 'b' / 'b2.TXT')),
        str((tmp_path / 'a' / 'b' / 'c' / 'c1.Txt')),
        str((tmp_path / 'a' / 'b' / 'c' / 'd' / 'd1.txt')),
        str((tmp_path / 'a' / 'b' / 'c' / 'd' / 'd2.txt')),
        str((tmp_path / 'foo.txt')),
    )
    assert fs.file_list(tmp_path, extensions=('jpg', 'txt')) == (
        str((tmp_path / 'a' / 'a1.txt')),
        str((tmp_path / 'a' / 'a2.jpg')),
        str((tmp_path / 'a' / 'b' / 'b1.JPG')),
        str((tmp_path / 'a' / 'b' / 'b2.TXT')),
        str((tmp_path / 'a' / 'b' / 'c' / 'c1.Txt')),
        str((tmp_path / 'a' / 'b' / 'c' / 'c2.jPG')),
        str((tmp_path / 'a' / 'b' / 'c' / 'd' / 'd1.txt')),
        str((tmp_path / 'a' / 'b' / 'c' / 'd' / 'd2.txt')),
        str((tmp_path / 'bar.jpg')),
        str((tmp_path / 'foo.txt')),
    )

def test_file_list_filters_by_age(tmp_path):
    def write(filepath, age):
        filepath.parent.mkdir(parents=True, exist_ok=True)
        filepath.write_bytes(b'foo')
        atime = time.time() - age
        os.utime(filepath, times=(atime, atime))

    write(tmp_path / 'a', 21)
    write(tmp_path / 'b', 20)
    write(tmp_path / 'c', 10)
    write(tmp_path / 'd', 9)
    write(tmp_path / '1' / 'a', 22)
    write(tmp_path / '1' / 'b', 19)
    write(tmp_path / '1' / 'c', 11)
    write(tmp_path / '1' / 'd', 8)
    write(tmp_path / '1' / '2' / 'a', 23)
    write(tmp_path / '1' / '2' / 'b', 18)
    write(tmp_path / '1' / '2' / 'c', 12)
    write(tmp_path / '1' / '2' / 'd', 7)
    files = fs.file_list(tmp_path, min_age=10, max_age=20)
    assert files == (
        str(tmp_path / '1' / '2' / 'b'),
        str(tmp_path / '1' / '2' / 'c'),
        str(tmp_path / '1' / 'b'),
        str(tmp_path / '1' / 'c'),
        str(tmp_path / 'b'),
        str(tmp_path / 'c'),
    )

@pytest.mark.parametrize('path', ('', None, 0))
def test_file_list_if_path_is_falsy(path):
    assert fs.file_list(path, extensions=('txt',)) == ()

def test_file_list_if_path_is_matching_nondirectory(tmp_path):
    path = tmp_path / 'foo.txt'
    path.write_bytes(b'foo')
    assert fs.file_list(path, extensions=('txt',)) == (str(path),)

def test_file_list_if_path_is_nonmatching_nondirectory(tmp_path):
    path = tmp_path / 'foo.txt'
    path.write_bytes(b'foo')
    assert fs.file_list(path, extensions=('png',)) == ()

def test_file_list_with_unreadable_subdirectory(tmp_path):
    (tmp_path / 'foo').write_bytes(b'foo')
    (tmp_path / 'a').mkdir()
    (tmp_path / 'a' / 'a.txt').write_bytes(b'foo')
    (tmp_path / 'a' / 'b').mkdir()
    (tmp_path / 'a' / 'b' / 'b.txt').write_bytes(b'foo')
    os.chmod(tmp_path / 'a' / 'b', 0o000)
    try:
        assert fs.file_list(tmp_path) == (
            str((tmp_path / 'a' / 'a.txt')),
            str((tmp_path / 'foo')),
        )
    finally:
        os.chmod(tmp_path / 'a' / 'b', 0o700)

def test_file_list_with_nonexisting_path(tmp_path):
    assert fs.file_list(tmp_path / 'does' / 'not' / 'exist') == (
        str(tmp_path / 'does' / 'not' / 'exist'),
    )


@pytest.mark.parametrize(
    argnames='name, path, tree, validator, exp_return_value',
    argvalues=(
        (
            'foo',
            'some/foo/path',
            (
                'a/asdf',
                'a/foo',
                'a/bsdf',
                'b/bar/foo/x',
                'c/baz/x/foo',
            ),
            None,
            '{tmp_path}/some/foo/path/a/foo',
        ),
        (
            'foo',
            'some/foo/path',
            (
                'a/asdf',
                'a/foo',
                'a/bsdf',
                'b/bar/foo/x',
                'c/baz/x/foo',
            ),
            os.path.isfile,
            '{tmp_path}/some/foo/path/a/foo',
        ),
        (
            'foo',
            'some/foo/path',
            (
                'a/asdf',
                'a/foo',
                'a/bsdf',
                'b/bar/foo/x',
                'c/baz/x/foo',
            ),
            os.path.isdir,
            '{tmp_path}/some/foo/path/b/bar/foo',
        ),
    ),
)
def test_find_name(name, path, tree, validator, exp_return_value, tmp_path):
    path = tmp_path / path
    path.mkdir(exist_ok=True, parents=True)
    tree = list(tree)
    random.shuffle(tree)
    for relfilepath in tree:
        (path / relfilepath).parent.mkdir(exist_ok=True, parents=True)
        (path / relfilepath).write_bytes(b'data')

    return_value = fs.find_name(name, str(path), validator=validator)
    assert return_value == exp_return_value.format(tmp_path=tmp_path)


def test_format_file_tree():
    tree = (
        ('root', (
            ('sub1', (
                ('foo', 100),
                ('sub2', (
                    ('sub3', (
                        ('sub4.1', (
                            ('bar1', 2000),
                        )),
                        ('sub4.2', (
                            ('bar2', 3500),
                        )),
                    )),
                )),
                ('baz', 5000),
            )),
        )),
    )

    assert fs.format_file_tree(tree) == '''
root (10.6 kB)
└─sub1 (10.6 kB)
  ├─foo (100 B)
  ├─sub2 (5.5 kB)
  │ └─sub3 (5.5 kB)
  │   ├─sub4.1 (2 kB)
  │   │ └─bar1 (2 kB)
  │   └─sub4.2 (3.5 kB)
  │     └─bar2 (3.5 kB)
  └─baz (5 kB)
'''.strip()



"""
Create screenshots from video file(s)
"""

import collections
import os
import queue

from .. import errors
from ..utils import LazyModule, daemon, fs, image, timestamp, torrent, video
from . import JobBase

import logging  # isort:skip
_log = logging.getLogger(__name__)

DEFAULT_NUMBER_OF_SCREENSHOTS = 2

natsort = LazyModule(module='natsort', namespace=globals())


class ScreenshotsJob(JobBase):
    """
    Create screenshots from video file(s)

    This job adds the following signals to the :attr:`~.JobBase.signal`
    attribute:

        ``screenshots_total``
            Emitted before screenshots are created. Registered callbacks get the
            total number of screenshots as a positional argument.
    """

    name = 'screenshots'
    label = 'Screenshots'
    cache_id = None

    def initialize(self, *, content_path, exclude_files=(), timestamps=(), count=0,
                   from_all_videos=False, optimize='default'):
        """
        Set internal state

        :param str content_path: Path to file or directory or sequence of paths
        :param exclude_files: Sequence of glob patterns (:class:`str`) and
            :class:`re.Pattern` objects (return value from :func:`re.compile`)
            that are matched against the relative path beneath `content_path`

            Glob patterns are matched case-insensitively.

            .. note:: Non-video files and stuff like `Sample.mkv` are always
                      excluded by :func:`.video.filter_main_videos`.
        :param timestamps: Screenshot positions in the video
        :type timestamps: sequence of "[[H+:]M+:]S+" strings or seconds
        :param count: How many screenshots to make
        :param bool from_all_videos: Whether to take `count` screenshots from
            each video file or only from the first video

            See :func:`.video.find_videos` for more information.

        :param optimize: `level` argument for :func:`~image.optimize`

            If this is ``"default"``, missing optimization dependencies are
            silently ignored.

        If `timestamps` and `count` are not given, screenshot positions are
        picked at even intervals. If `count` is larger than the number of
        `timestamps`, more timestamps are added.
        """
        self._content_path = content_path
        self._exclude_files = exclude_files
        self._screenshots_created = 0
        self._screenshots_total = -1
        self._timestamps = timestamps
        self._count = count
        self._from_all_videos = from_all_videos
        self._optimize = optimize
        self._screenshots_process = None
        self._optimize_process = None
        self._screenshots_by_file = collections.defaultdict(lambda: [])
        self.signal.add('screenshots_total', record=True)

    def execute(self):
        """
        Execute screenshot creation subprocess

        The subprocess also picks video file(s) and timestamps.
        """
        if self._optimize not in ('none', None):
            self._execute_screenshots_process()
            self._execute_optimize_process()
        else:
            self._execute_screenshots_process()

    def _execute_screenshots_process(self):
        self._screenshots_process = daemon.DaemonProcess(
            name='_screenshots_process',
            target=_screenshots_process,
            kwargs={
                'content_path': self._content_path,
                'exclude_files': self._exclude_files,
                'timestamps': self._timestamps,
                'count': self._count,
                'from_all_videos': self._from_all_videos,
                'output_dir': self.home_directory,
                'overwrite': self.ignore_cache,
            },
            info_callback=self._handle_info,
            error_callback=self._handle_error,
        )
        self._screenshots_process.start()

    def _execute_optimize_process(self):
        self._optimize_process = daemon.DaemonProcess(
            name='_optimize_process',
            target=_optimize_process,
            kwargs={
                # Use default optimization level if None is provided
                'level': self._optimize,
                # Ignore missing dependecy if we do "default" optimization
                'ignore_dependency_error': (
                    True
                    if self._optimize == 'default' else
                    False
                ),
            },
            info_callback=self._handle_info,
            error_callback=self._handle_error,
        )
        self._optimize_process.start()

    def _handle_info(self, info):
        if not self.is_finished:
            if 'screenshots_total' in info:
                self._screenshots_total = info['screenshots_total']
                self.signal.emit('screenshots_total', self._screenshots_total)

            if 'screenshot_path' in info:
                self._screenshots_by_file[info['video_path']].append(info['screenshot_path'])

                if self._optimize_process:
                    _log.debug('Optimizing %s: %.2f KiB',
                               info['screenshot_path'],
                               fs.file_size(info['screenshot_path']) / 1024)
                    self._optimize_process.send(daemon.MsgType.info, info['screenshot_path'])

                else:
                    self._screenshots_created += 1
                    self.send(info['screenshot_path'])

            elif 'optimized_screenshot_path' in info:
                _log.debug('Optimized %s: %.2f KiB',
                           info['optimized_screenshot_path'],
                           fs.file_size(info['optimized_screenshot_path']) / 1024)
                self._screenshots_created += 1
                self.send(info['optimized_screenshot_path'])

            if self.screenshots_created == self.screenshots_total:
                self.finish()

    def _handle_error(self, error):
        if isinstance(error, BaseException):
            self.exception(error)
        else:
            self.error(error)

    def finish(self):
        """Terminate screenshot creation subprocess and finish"""
        if self._screenshots_process:
            self._screenshots_process.stop()
        if self._optimize_process:
            self._optimize_process.stop()
        super().finish()

    @property
    def exit_code(self):
        """`0` if all screenshots were made, `1` otherwise, `None` if unfinished"""
        if self.is_finished:
            if self.screenshots_total < 0:
                # Job is finished but _screenshots_process() never sent us
                # timestamps. That means we're either using previously cached
                # output or the job was cancelled while _screenshots_process()
                # was still initializing.
                if self.output:
                    # If we have cached output, assume the cached number of
                    # screenshots is what the user wanted because the output of
                    # unsuccessful jobs is not cached (see
                    # JobBase._write_output_cache()).
                    return 0
                else:
                    return 1
            elif len(self.output) == self.screenshots_total:
                return 0
            else:
                return 1

    @property
    def content_path(self):
        """
        Path or sequence of file paths

        If this is set to a single path, appropriate files are picked by
        :func:`~.video.find_videos`.

        If this is set to a sequence of file paths, non-video files and uncommon
        playback durations are excluded. (See
        :func:`~.video.filter_main_videos`.)

        See also :attr:`from_all_videos`.

        Setting this property when this job :attr:`~.JobBase.is_started` raises
        :class:`RuntimeError`.
        """
        return self._content_path

    @content_path.setter
    def content_path(self, content_path):
        if self.is_started:
            raise RuntimeError('Cannot set content_path after job has been started')
        else:
            self._content_path = content_path

    @property
    def exclude_files(self):
        """
        Sequence of glob and :class:`regex <re.Pattern>` patterns to exclude

        See :meth:`initialize` for more information.

        Setting this property when this job :attr:`~.JobBase.is_started` raises
        :class:`RuntimeError`.
        """
        return self._exclude_files

    @exclude_files.setter
    def exclude_files(self, exclude_files):
        if self.is_started:
            raise RuntimeError('Cannot set exclude_files after job has been started')
        else:
            self._exclude_files = exclude_files

    @property
    def from_all_videos(self):
        """
        Whether to make screenshots from all video files or only the first

        Setting this property when this job :attr:`~.JobBase.is_started` raises
        :class:`RuntimeError`.
        """
        return self._from_all_videos

    @from_all_videos.setter
    def from_all_videos(self, from_all_videos):
        if self.is_started:
            raise RuntimeError('Cannot set from_all_videos after job has been started')
        else:
            self._from_all_videos = from_all_videos

    @property
    def count(self):
        """
        How many screenshots to make per video file

        Setting this property when this job :attr:`~.JobBase.is_started` raises
        :class:`RuntimeError`.
        """
        return self._count

    @count.setter
    def count(self, count):
        if self.is_started:
            raise RuntimeError('Cannot set count after job has been started')
        else:
            self._count = count

    @property
    def timestamps(self):
        """
        Specific list of timestamps to make

        Setting this property when this job :attr:`~.JobBase.is_started` raises
        :class:`RuntimeError`.
        """
        return self._timestamps

    @timestamps.setter
    def timestamps(self, timestamps):
        if self.is_started:
            raise RuntimeError('Cannot set timestamps after job has been started')
        else:
            self._timestamps = timestamps

    @property
    def screenshots_total(self):
        """
        Total number of screenshots to make

        .. note:: This is ``-1`` until the subprocess that creates the
                  screenshots is executed and determined the number of
                  screenshots.
        """
        return self._screenshots_total

    @property
    def screenshots_created(self):
        """Total number of screenshots made so far"""
        return self._screenshots_created

    @property
    def screenshots_by_file(self):
        """
        Map source video file paths to sequence of generated screenshot file
        paths so far
        """
        return {
            video_path: tuple(screenshot_paths)
            for video_path, screenshot_paths in self._screenshots_by_file.items()
        }


def _screenshots_process(output_queue, input_queue,
                         *, content_path, exclude_files, timestamps, count, from_all_videos,
                         output_dir, overwrite):
    video_files = list(_get_video_files(
        output_queue,
        content_path=content_path,
        exclude_files=exclude_files,
    ))
    if not video_files:
        output_queue.put((daemon.MsgType.error, 'No videos found'))
        return

    if not from_all_videos:
        # Only create screenshots from first video
        del video_files[1:]

    timestamps_map = _map_timestamps(
        video_files=video_files,
        timestamps=timestamps,
        count=count,
    )

    # Herald how many screenshots we are going to make
    screenshots_total = sum((len(ts) for ts in timestamps_map.values()))
    output_queue.put((daemon.MsgType.info, {'screenshots_total': screenshots_total}))

    try:
        _screenshot_video_files(
            output_queue, input_queue,
            timestamps_map=timestamps_map,
            output_dir=output_dir,
            overwrite=overwrite,
        )
    except SystemExit:
        # _maybe_terminate signals termination by raising SystemExit. This
        # shouldn't cause any issues as long as we actually exit here.
        pass


def _optimize_process(output_queue, input_queue, *, level, ignore_dependency_error):
    # Get as many queued screenshots as possible. Check if there's a termination
    # sentinel before optimizing each screenshot to avoid processing all the
    # screenshots before Ctrl-C takes effect.
    msgs = []
    while True:
        new_msgs = _read_queue_until_empty(input_queue=input_queue)
        msgs.extend(new_msgs)

        if any(typ is daemon.MsgType.terminate for typ, msg_ in msgs):
            break

        elif msgs:
            typ_, screenshot_file = msgs.pop(0)
            try:
                image.optimize(screenshot_file, level=level)

            except errors.ImageOptimizeError as e:
                output_queue.put((daemon.MsgType.error, str(e)))

            except errors.DependencyError as e:
                if ignore_dependency_error:
                    # Act like we optimized `screenshot_file`
                    output_queue.put((daemon.MsgType.info, {
                        'optimized_screenshot_path': screenshot_file,
                    }))
                else:
                    raise e

            else:
                output_queue.put((daemon.MsgType.info, {
                    'optimized_screenshot_path': screenshot_file,
                }))


def _get_video_files(output_queue, *, content_path, exclude_files):
    # Make list of appropriate video file(s) from `content_path`
    try:
        filtered_files = torrent.filter_files(content_path, exclude=exclude_files)
        return video.filter_main_videos(filtered_files)
    except errors.ContentError as e:
        output_queue.put((daemon.MsgType.error, str(e)))
        return []


def _screenshot_video_files(output_queue, input_queue, *, timestamps_map, output_dir, overwrite):
    # Make all screenshots from all video files
    for video_file, timestamps in timestamps_map.items():
        _maybe_terminate(input_queue=input_queue)
        _screenshot_video_file(
            output_queue, input_queue,
            video_file=video_file,
            timestamps=timestamps,
            output_dir=output_dir,
            overwrite=overwrite,
        )


def _screenshot_video_file(output_queue, input_queue, *, video_file, timestamps, output_dir, overwrite):
    # Make all screenshots from one video file
    for ts in timestamps:
        _maybe_terminate(input_queue=input_queue)
        _make_screenshot(
            output_queue,
            video_file=video_file,
            timestamp=ts,
            output_dir=output_dir,
            overwrite=overwrite,
        )


def _make_screenshot(output_queue, *, video_file, timestamp, output_dir, overwrite):
    # Make one screenshot from one video file
    screenshot_file = os.path.join(
        output_dir,
        fs.basename(video_file) + f'.{timestamp}.png',
    )
    try:
        actual_screenshot_file = image.screenshot(
            video_file=video_file,
            screenshot_file=screenshot_file,
            timestamp=timestamp,
            overwrite=overwrite,
        )
    except errors.ScreenshotError as e:
        output_queue.put((daemon.MsgType.error, str(e)))
    else:
        output_queue.put((daemon.MsgType.info, {
            'video_path': video_file,
            'screenshot_path': actual_screenshot_file,
        }))


def _map_timestamps(*, video_files, timestamps, count):
    # Map each video_file to a sequence of timestamps
    timestamps_map = {}
    for video_file in video_files:
        timestamps_map[video_file] = _validate_timestamps(
            video_file=video_file,
            timestamps=timestamps,
            count=count,
        )
    return timestamps_map


def _validate_timestamps(*, video_file, timestamps, count):
    # Validate, normalize, deduplicate and sort timestamps

    # Stay clear of the last 10 seconds
    duration = video.duration(video_file) - 10
    if duration < 60:
        return []

    timestamps_pretty = set()
    for ts in timestamps:
        ts = max(0, min(duration, timestamp.parse(ts)))
        timestamps_pretty.add(timestamp.pretty(ts))

    if not timestamps and not count:
        count = DEFAULT_NUMBER_OF_SCREENSHOTS

    # Add more timestamps if the user didn't specify enough
    if count > 0 and len(timestamps_pretty) < count:
        # Convert timestamp strings to seconds
        timestamps = sorted(timestamp.parse(ts) for ts in timestamps_pretty)

        # Fractions of video duration
        positions = [ts / duration for ts in timestamps]

        # Include start and end of video
        if 0.0 not in positions:
            positions.insert(0, 0.0)
        if 1.0 not in positions:
            positions.append(1.0)

        # Insert timestamps between the two positions with the largest distance
        while len(timestamps_pretty) < count:
            pairs = [(a, b) for a, b in zip(positions, positions[1:])]
            max_distance, pos1, pos2 = max((b - a, a, b) for a, b in pairs)
            position = ((pos2 - pos1) / 2) + pos1
            timestamps_pretty.add(timestamp.pretty(duration * position))
            positions.append(position)
            positions.sort()

    return natsort.natsorted(timestamps_pretty)


def _maybe_terminate(*, input_queue):
    try:
        typ, msg = input_queue.get_nowait()
    except queue.Empty:
        pass
    else:
        if typ == daemon.MsgType.terminate:
            raise SystemExit('Terminated')


def _read_queue_until_empty(*, input_queue):
    msgs = []
    while True:
        try:
            typ, msg = input_queue.get(timeout=0.01)
        except queue.Empty:
            return msgs
        else:
            msgs.append((typ, msg))



from itertools import zip_longest

import pytest

from upsies.utils.types import ReleaseType
from upsies.utils.webdbs import Query, SearchResult, imdb


@pytest.fixture
def api():
    return imdb.ImdbApi()


@pytest.mark.parametrize(
    argnames='query, exp_query',
    argvalues=(
        (Query('Foo and Bar'), Query('foo bar')),
        (Query('Foo & Bar'), Query('foo & bar')),
        (Query('Cant See You'), Query("can't See You")),
        (Query('Wont See You'), Query("Won't See You")),
        (Query('Dont See You'), Query("Don't See You")),
        (Query('Want to See You'), Query("Want to See You")),
    ),
    ids=lambda value: str(value),
)
def test_sanitize_query(query, exp_query, api, store_response):
    return_value = api.sanitize_query(query)
    assert return_value == exp_query


@pytest.mark.parametrize(
    argnames='text, exp_id',
    argvalues=(
        ('Some Title', None),
        ('1984', None),
        ('tt0123arf', None),
        ('arftt0123', None),
        ('tt0123', 'tt0123'),
        ('https://www.imdb.com/title/tt0123/', 'tt0123'),
        ('https://www.imdb.com/title/tt0123', 'tt0123'),
        ('junk tt0123 more junk', 'tt0123'),
        ('junk https://www.imdb.com/title/tt0123/stuff more junk', 'tt0123'),
    ),
    ids=lambda value: str(value),
)
def test_get_id_from_text(text, exp_id, api, store_response):
    return_value = api.get_id_from_text(text)
    assert return_value == exp_id


@pytest.mark.parametrize(
    argnames='id, title, title_english, title_original, type, year, cast, countries, directors, genres, summary',
    argvalues=(
        (
            'tt3286052',
            "The Blackcoat's Daughter",
            "The Blackcoat's Daughter",
            'February',
            ReleaseType.movie,
            '2015',
            ('Emma Roberts', 'Kiernan Shipka', 'Lucy Boynton'),
            ('Canada', 'United States'),
            ('Oz Perkins',),
            ('horror', 'mystery', 'thriller'),
            ('Two girls must battle a mysterious evil force when they get left behind '
             'at their boarding school over winter break.'),
        ),
        (
            'tt0080455',
            'The Blues Brothers',
            '',
            'The Blues Brothers',
            ReleaseType.movie,
            '1980',
            ('John Belushi', 'Dan Aykroyd', 'Cab Calloway'),
            ('United States',),
            ('John Landis',),
            ('action', 'adventure', 'comedy'),
            ('Jake Blues rejoins with his brother Elwood after being released from prison, '
             'but the duo has just days to reunite their old R&B band and save the Catholic '
             'home where the two were raised, outrunning the police as they tear through Chicago.'),
        ),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_handles_id_in_query(id, title, title_english, title_original, type, year,
                                          cast, countries, directors, genres, summary,
                                          api, store_response):
    results = await api.search(Query(id=id))
    assert len(results) == 1
    assert results[0].id == id
    assert results[0].title == title
    assert results[0].type == type
    assert results[0].url == f'{imdb.ImdbApi._url_base}/title/{id}'
    assert results[0].year == year

    # Make sure we're not awaiting the same coroutine again on subsequent calls
    for _ in range(2):
        assert (await results[0].cast())[:3] == cast
        assert (await results[0].countries()) == countries
        assert (await results[0].directors()) == directors
        assert (await results[0].genres()) == genres
        assert (await results[0].summary()) == summary
        assert (await results[0].title_english()) == title_english
        assert (await results[0].title_original()) == title_original


@pytest.mark.asyncio
async def test_search_returns_empty_list_if_title_is_empty(api, store_response):
    assert await api.search(Query('', year='2009')) == []

@pytest.mark.asyncio
async def test_search_returns_list_of_SearchResults(api, store_response):
    results = await api.search(Query('Star Wars'))
    for result in results:
        assert isinstance(result, SearchResult)


@pytest.mark.parametrize(
    argnames=('query', 'exp_title'),
    argvalues=(
        (Query('Blues Brothers', year=1980), 'The Blues Brothers'),
        (Query('Blues Brothers', year='1998'), 'Blues Brothers 2000'),
        (Query('Blues Brothers', year=1950), ''),
        (Query('Star Wars', year=2008), 'Star Wars: The Clone Wars'),
        (Query('Star Wars', year=2014), 'Star Wars: Rebels'),
        (Query('Star Wars', year=1950), ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_for_year(query, exp_title, api, store_response):
    results = await api.search(query)
    if exp_title:
        assert results[0].title == exp_title
    else:
        assert not results


@pytest.mark.parametrize(
    argnames=('query', 'exp_titles'),
    argvalues=(
        (Query('star wars', type=ReleaseType.series), ('Star Wars: Clone Wars', 'Star Wars: The Clone Wars')),
        (Query('rampage', type=ReleaseType.series), ('Road to Rampage', 'Rampage: Killing Without Reason')),
        (Query('balada triste trompeta', type=ReleaseType.series), ()),
        (Query('time traveling bong', type=ReleaseType.series), ('Time Traveling Bong',)),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_for_series(query, exp_titles, api, store_response):
    results = await api.search(query)
    titles = [r.title for r in results]
    if exp_titles:
        for exp_title in exp_titles:
            assert exp_title in titles
    else:
        assert not titles


@pytest.mark.parametrize(
    argnames=('query', 'exp_titles'),
    argvalues=(
        (Query('blues brothers', type=ReleaseType.movie), ('The Blue Brothers', 'Blues Brothers 2000', 'Brothers in Blue')),
        (Query('rampage', type=ReleaseType.movie), ('Rampage: Capital Punishment', 'American Rampage')),
        (Query('balada triste trompeta', type=ReleaseType.movie), ('The Last Circus',)),
        (Query('time traveling bong', type=ReleaseType.movie), ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_for_movie(query, exp_titles, api, store_response):
    results = await api.search(query)
    titles = [r.title for r in results]
    if exp_titles:
        for exp_title in exp_titles:
            assert exp_title in titles
    else:
        assert not titles


@pytest.mark.parametrize(
    argnames=('query', 'exp_cast'),
    argvalues=(
        (Query('The Blues Brothers', year=1980), ('Dan Aykroyd', 'John Belushi')),
        (Query('Deadwood', year=2004), ('Brad Dourif', 'Ian McShane')),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_cast(query, exp_cast, api, store_response):
    results = await api.search(query)
    cast = await results[0].cast()
    for member in exp_cast:
        assert member in cast


@pytest.mark.parametrize(
    argnames=('query', 'exp_countries'),
    argvalues=(
        (Query('Blues Brothers', type=ReleaseType.movie, year=1980), ('United States',)),
        (Query('Bron Broen', type=ReleaseType.series, year=2011), ('Sweden', 'Denmark', 'Germany', 'Norway')),
        (Query('The Forest', type=ReleaseType.series, year=2017), ('France',)),
        (Query('Karppi', type=ReleaseType.series, year=2018), ('Finland', 'Germany', 'France')),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_country(query, exp_countries, api, store_response):
    results = await api.search(query)
    countries = await results[0].countries()
    assert countries == exp_countries


@pytest.mark.parametrize(
    argnames=('query', 'exp_id'),
    argvalues=(
        (Query('Blues Brothers', year=1980), 'tt0080455'),
        (Query('Blues Brothers', year=1998), 'tt0118747'),
        (Query('Star Wars', year=2008), 'tt0458290'),
        (Query('Star Wars', year=2014), 'tt2930604'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_id(query, exp_id, api, store_response):
    results = await api.search(query)
    assert results[0].id == exp_id


@pytest.mark.parametrize(
    argnames=('query', 'exp_directors'),
    argvalues=(
        (Query('star wars', type=ReleaseType.movie, year=1977), ('George Lucas',)),
        (Query('balada triste trompeta', type=ReleaseType.movie, year=2010), ('Álex de la Iglesia',)),
        (Query('The Forest', type=ReleaseType.series, year=2017), ()),
        (Query('Karppi', type=ReleaseType.series, year=2018), ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_directors(query, exp_directors, api, store_response):
    results = await api.search(query)
    directors = await results[0].directors()
    assert directors == exp_directors


@pytest.mark.parametrize(
    argnames=('query', 'exp_genres'),
    argvalues=(
        (Query('star wars', type=ReleaseType.movie, year=1977), ('action', 'adventure', 'fantasy')),
        (Query('balada triste trompeta', type=ReleaseType.movie, year=2010), ('adventure', 'comedy', 'drama')),
        (Query('The Forest', type=ReleaseType.series, year=2017), ('crime', 'drama')),
        (Query('Deadwood', type=ReleaseType.series, year=2004), ('crime', 'drama', 'history')),
        (Query('The Believer', type=ReleaseType.movie, year=2001), ('drama',)),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_genres(query, exp_genres, api, store_response):
    results = await api.search(query)
    genres = await results[0].genres()
    assert genres == exp_genres


@pytest.mark.parametrize(
    argnames=('query', 'exp_summary'),
    argvalues=(
        (Query('star wars', type=ReleaseType.movie, year=1977),
         ("Luke Skywalker joins forces with a Jedi Knight, a cocky pilot, a Wookiee and two droids to save "
          "the galaxy from the Empire's world-destroying battle station, while also attempting to rescue "
          "Princess Leia from the mysterious Darth Vader.")),
        (Query('balada triste trompeta', type=ReleaseType.movie, year=2010),
         ("A young trapeze artist must decide between her lust for Sergio, the Happy Clown, "
          "or her affection for Javier, the Sad Clown, both of whom are deeply disturbed.")),
        (Query('The Forest', type=ReleaseType.series, year=2017),
         ("Sixteen-year-old Jennifer disappears one night from her village in the Ardennes. "
          "Captain Gaspard Deker leads the investigation with local cop Virginie Musso, "
          "who knew the girl well. They are helped by Eve, a lonely and mysterious woman.")),
        (Query('Deadwood', type=ReleaseType.series, year=2004),
         ("A show set in the late 1800s, revolving around the characters of Deadwood, "
          "South Dakota; a town of deep corruption and crime.")),
        (Query('The Deadwood Coach', type=ReleaseType.movie, year=1924), ''),
        (Query('Two Down', type=ReleaseType.movie, year=2015),
         ("A young woman's world is turned upside down when an injured hit man "
          "takes her hostage at gunpoint in her own home.")),
        (Query('My Best Fiend', type=ReleaseType.movie, year=1999),
         ("In the 1950s, an adolescent Werner Herzog was transfixed by a film performance "
          "of the young Klaus Kinski. Years later, they would share an apartment where, "
          "in an unabated, forty-eight-hour…")),
        (Query('Zero Dark Thirty', type=ReleaseType.movie, year=2012),
         ("A chronicle of the decade-long hunt for al-Qaeda terrorist leader "
          "Osama bin Laden after the September 2001 attacks, and his death "
          "at the hands of the Navy S.E.A.L.s Team 6 in May 2011.")),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_summary(query, exp_summary, api, store_response):
    results = await api.search(query)
    summary = await results[0].summary()
    assert summary == exp_summary

@pytest.mark.parametrize(
    argnames=('query', 'unexp_summary'),
    argvalues=(
        (Query('the tick', year=1994), 'TV-Y7'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_summary_negated(query, unexp_summary, api, store_response):
    results = await api.search(query)
    assert results
    for r in results:
        summary = await r.summary()
        assert unexp_summary not in summary


@pytest.mark.parametrize(
    argnames=('query', 'exp_title'),
    argvalues=(
        (Query('Blues Brothers', year=1980), 'The Blues Brothers'),
        (Query('Blues Brothers', year=1998), 'Blues Brothers 2000'),
        (Query('Star Wars', year=2008), 'Star Wars: The Clone Wars'),
        (Query('Star Wars', year=2014), 'Star Wars: Rebels'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_title(query, exp_title, api, store_response):
    results = await api.search(query)
    assert results[0].title == exp_title


@pytest.mark.parametrize(
    argnames=('query', 'exp_title_english', 'exp_title_original'),
    argvalues=(
        (Query('36th Chamber of Shaolin', type=ReleaseType.movie, year=1978), 'The 36th Chamber of Shaolin', 'Shao Lin san shi liu fang'),
        (Query('Aftermath', type=ReleaseType.movie, year=2012), 'Aftermath', 'Poklosie'),
        (Query('balada triste trompeta', type=ReleaseType.movie, year=2010), 'The Last Circus', 'Balada triste de trompeta'),
        (Query('Blues Brothers', type=ReleaseType.movie, year=1980), '', 'The Blues Brothers'),
        (Query('Cyborg', type=ReleaseType.movie, year=1989), '', 'Cyborg'),
        (Query('Deadwood', type=ReleaseType.series, year=2004), '', 'Deadwood'),
        (Query('February', type=ReleaseType.movie, year=2015), "The Blackcoat's Daughter", 'February'),
        (Query('Hard Boiled', type=ReleaseType.movie, year=1992), 'Hard Boiled', 'Lat sau san taam'),
        (Query('Karppi', type=ReleaseType.series, year=2018), 'Deadwind', 'Karppi'),
        (Query('Olympus Has Fallen', type=ReleaseType.movie, year=2013), '', 'Olympus Has Fallen'),
        (Query('Sin Nombre', type=ReleaseType.movie, year=2009), '', 'Sin nombre'),
        (Query('Stone Cold', type=ReleaseType.movie, year=1991), '', 'Stone Cold'),
        (Query('The Forest', type=ReleaseType.series, year=2017), 'The Forest', 'La forêt'),
        (Query('The Nest', type=ReleaseType.movie, year=2002), 'The Nest', 'Nid de guêpes'),
        (Query('Traffic Light', type=ReleaseType.series, year=2008), 'Traffic Light', 'Ramzor'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_title_english_original(query, exp_title_english, exp_title_original, api, store_response):
    results = await api.search(query)
    title_english = await results[0].title_english()
    title_original = await results[0].title_original()
    assert title_english == exp_title_english
    assert title_original == exp_title_original


@pytest.mark.parametrize(
    argnames=('query', 'exp_type'),
    argvalues=(
        (Query('Blues Brothers', year=1980), ReleaseType.movie),
        (Query('Blues Brothers', year=1998), ReleaseType.movie),
        (Query('Star Wars', year=2008), ReleaseType.series),
        (Query('Star Wars', year=2014), ReleaseType.series),
        (Query('Elephant', year=1989, type=ReleaseType.movie), ReleaseType.movie),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_type(query, exp_type, api, store_response):
    results = await api.search(query)
    assert results[0].type == exp_type


@pytest.mark.parametrize(
    argnames=('query', 'exp_url'),
    argvalues=(
        (Query('The Blues Brothers', year=1980), 'https://www.imdb.com/title/tt0080455'),
        (Query('Deadwood', year=2004), 'https://www.imdb.com/title/tt0348914'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_url(query, exp_url, api, store_response):
    results = await api.search(query)
    assert results[0].url == exp_url


@pytest.mark.parametrize(
    argnames=('query', 'exp_year'),
    argvalues=(
        (Query('The Blues Brothers', year=1980), '1980'),
        (Query('Deadwood', year=2004), '2004'),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_search_result_year(query, exp_year, api, store_response):
    results = await api.search(query)
    assert results[0].year == exp_year

@pytest.mark.asyncio
async def test_search_result_parser_failure(api):
    result = imdb._ImdbSearchResult(imdb_api=api)
    assert result.id == ''
    assert result.title == ''
    assert result.type == ReleaseType.series
    assert result.url == ''
    assert result.year == ''
    assert await result.cast() == ()
    assert await result.countries() == ()
    assert await result.directors() == ()
    assert await result.genres() == ()
    assert await result.summary() == ''
    assert await result.title_english() == ''
    assert await result.title_original() == ''


@pytest.mark.parametrize(
    argnames=('id', 'exp_cast'),
    argvalues=(
        # Blues Brothers (movie)
        ('tt0080455', (('John Belushi', 'https://www.imdb.com/name/nm0000004'),
                       ('Dan Aykroyd', 'https://www.imdb.com/name/nm0000101'),
                       ('Cab Calloway', 'https://www.imdb.com/name/nm0130572'))),
        # Wind in the Willows (TV movie)
        ('tt0192802', (('Alan Bennett', 'https://www.imdb.com/name/nm0003141'),
                       ('Michael Palin', 'https://www.imdb.com/name/nm0001589'),
                       ('Michael Gambon', 'https://www.imdb.com/name/nm0002091'))),
        # Bender's Big Score (Video)
        ('tt0471711', (('Billy West', 'https://www.imdb.com/name/nm0921942'),
                       ('Katey Sagal', 'https://www.imdb.com/name/nm0005408'),
                       ('John DiMaggio', 'https://www.imdb.com/name/nm0224007'))),
        # Elephant (TV Short)
        ('tt0097270', (('Gary Walker', 'https://www.imdb.com/name/nm1281884'),
                       ('Bill Hamilton', 'https://www.imdb.com/name/nm1685093'),
                       ('Michael Foyle', 'https://www.imdb.com/name/nm0289432'))),
        # Kung Fury (Short)
        ('tt3472226', (('David Sandberg', 'https://www.imdb.com/name/nm6247887'),
                       ('Jorma Taccone', 'https://www.imdb.com/name/nm1672246'),
                       ('Steven Chew', 'https://www.imdb.com/name/nm7320022'))),
        # The Forest (TV mini-series)
        ('tt6560040', (('Samuel Labarthe', 'https://www.imdb.com/name/nm0479355'),
                       ('Suzanne Clément', 'https://www.imdb.com/name/nm0167501'),
                       ('Alexia Barlier', 'https://www.imdb.com/name/nm1715145'))),
        # Orange Is the New Black (series)
        ('tt2372162', (('Taylor Schilling', 'https://www.imdb.com/name/nm2279940'),
                       ('Danielle Brooks', 'https://www.imdb.com/name/nm5335029'),
                       ('Taryn Manning', 'https://www.imdb.com/name/nm0543383'))),
        # Orange Is the New Black - S07E01 (episode)
        ('tt5440238', (('Taylor Schilling', 'https://www.imdb.com/name/nm2279940'),
                       ('Natasha Lyonne', 'https://www.imdb.com/name/nm0005169'),
                       ('Uzo Aduba', 'https://www.imdb.com/name/nm2499064'))),
        # No cast list
        ('tt0896516', ()),
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_cast(id, exp_cast, api, store_response):
    cast = (await api.cast(id))[:3]
    if not cast:
        assert exp_cast == ()
    else:
        for person, (name, url) in zip_longest(cast, exp_cast):
            assert person == name
            assert person.url.rstrip('/') == url.rstrip('/')


@pytest.mark.parametrize(
    argnames=('id', 'exp_countries'),
    argvalues=(
        ('tt0080455', ('United States',)),  # Blues Brothers (movie)
        ('tt3286052', ('Canada', 'United States')),  # February (movie)
        ('tt0048918', ('United Kingdom',)),  # 1984 (movie)
        ('tt0192802', ('United Kingdom',)),  # Wind in the Willows (TV movie)
        ('tt0471711', ('United States',)),  # Bender's Big Score (Video)
        ('tt0097270', ('United Kingdom',)),  # Elephant (TV Short)
        ('tt3472226', ('Sweden',)),  # Kung Fury (Short)
        ('tt1733785', ('Sweden', 'Denmark', 'Germany', 'Norway')),  # The Bridge (series)
        ('tt0348914', ('United States',)),  # Deadwood (series)
        ('tt9319668', ('Germany', 'United States')),  # 1899 (series)
        ('tt0556307', ('United States',)),  # Deadwood - S02E04 (episode)
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_countries(id, exp_countries, api, store_response):
    countries = await api.countries(id)
    assert countries == exp_countries


@pytest.mark.parametrize(
    argnames=('id', 'exp_creators'),
    argvalues=(
        ('tt3286052', ()),  # February (movie)
        ('tt0192802', ()),  # Wind in the Willows (TV movie)
        ('tt0471711', ()),  # Bender's Big Score (Video)
        ('tt0097270', ()),  # Elephant (TV Short)
        ('tt3472226', ()),  # Kung Fury (Short)
        ('tt6560040', (('Delinda Jacobs', 'https://www.imdb.com/name/nm3064398'),)),  # The Forest (TV mini-series)
        ('tt2372162', (('Jenji Kohan', 'https://www.imdb.com/name/nm0463176'),)),  # Orange Is the New Black (series)
        ('tt0149460', (('David X. Cohen', 'https://www.imdb.com/name/nm0169326'),
                       ('Matt Groening', 'https://www.imdb.com/name/nm0004981'))),  # Futurama
        ('tt5440238', ()),  # Orange Is the New Black - S07E01 (episode)
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_creators(id, exp_creators, api, store_response):
    creators = await api.creators(id)
    if not creators:
        assert exp_creators == ()
    else:
        for person, (name, url) in zip_longest(creators, exp_creators, fillvalue=(None, None)):
            assert person == name
            assert person.url.rstrip('/') == url.rstrip('/')


@pytest.mark.parametrize(
    argnames=('id', 'exp_directors'),
    argvalues=(
        ('tt3286052', (('Oz Perkins', 'https://www.imdb.com/name/nm0674020'),)),  # February (movie)
        ('tt0192802', (('Dave Unwin', 'https://www.imdb.com/name/nm0881386'),
                       ('Dennis Abey', 'https://www.imdb.com/name/nm0008688'))),  # Wind in the Willows (TV movie)
        ('tt0471711', (('Dwayne Carey-Hill', 'https://www.imdb.com/name/nm1401752'),)),  # Bender's Big Score (Video)
        ('tt0097270', (('Alan Clarke', 'https://www.imdb.com/name/nm0164639'),)),  # Elephant (TV Short)
        ('tt3472226', (('David Sandberg', 'https://www.imdb.com/name/nm6247887'),)),  # Kung Fury (Short)
        ('tt6560040', ()),  # The Forest (TV mini-series)
        ('tt2372162', ()),  # Orange Is the New Black (series)
        ('tt5440238', (('Michael Trim', 'https://www.imdb.com/name/nm0872841'),)),  # Orange Is the New Black - S07E01 (episode)
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_directors(id, exp_directors, api, store_response):
    directors = await api.directors(id)
    if not directors:
        assert exp_directors == ()
    else:
        for person, (name, url) in zip_longest(directors[:3], exp_directors, fillvalue=(None, None)):
            assert person == name
            assert person.url.rstrip('/') == url.rstrip('/')


@pytest.mark.parametrize(
    argnames=('id', 'exp_genres'),
    argvalues=(
        ('tt0080455', ('action', 'adventure', 'comedy')),  # Blues Brothers (movie)
        ('tt0247199', ('drama',)),  # The Believer (single keyword)
        ('tt0192802', ('animation', 'family')),  # Wind in the Willows (TV movie)
        ('tt0471711', ('animation', 'comedy', 'musical')),  # Bender's Big Score (Video)
        ('tt0097270', ('crime', 'drama')),  # Elephant (TV movie)
        ('tt3472226', ('short', 'action', 'comedy')),  # Kung Fury (Short)
        ('tt6560040', ('crime', 'drama')),  # The Forest (mini series)
        ('tt0348914', ('crime', 'drama', 'history')),  # Deadwood (series)
        ('tt0556307', ('crime', 'drama', 'history')),  # Deadwood - S02E04 (episode)
        (None, ()),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_genres(id, exp_genres, api, store_response):
    genres = await api.genres(id)
    assert genres == exp_genres


@pytest.mark.parametrize(
    argnames='id, exp_poster_url',
    argvalues=(
        # Blues Brothers (movie)
        ('tt0080455', 'https://m.media-amazon.com/images/M/MV5BYTdlMDExOGUtN2I3MS00MjY5LWE1NTAtYzc3MzIxN2M3OWY1XkEyXkFqcGdeQXVyNzkwMjQ5NzM@._V1_UY1200_AL_.jpg'),
        # Wind in the Willows (TV movie)
        ('tt0192802', 'https://m.media-amazon.com/images/M/MV5BOGQ3NWM5OTAtOThjZS00MzMyLTkwZjMtOWIyOTdlZjFiZGI1XkEyXkFqcGdeQXVyNzMwOTY2NTI@._V1_UY1200_AL_.jpg'),
        # Bender's Big Score (Video)
        ('tt0471711', 'https://m.media-amazon.com/images/M/MV5BM2U3NmRiNDItMDQwYy00Y2Q0LWJiYzItNjAzNThkZjFlM2RiXkEyXkFqcGdeQXVyNTA4NzY1MzY@._V1_UY1200_AL_.jpg'),
        # Elephant (TV movie)
        ('tt0097270', 'https://m.media-amazon.com/images/M/MV5BMWE0ZTBiOWItY2ZkNS00MzI0LWE5Y2QtYjJmNjM1MGRkZGNmXkEyXkFqcGdeQXVyMzU0MTk1Nzc@._V1_UY1200_AL_.jpg'),
        # Kung Fury (Short)
        ('tt3472226', 'https://m.media-amazon.com/images/M/MV5BMjQwMjU2ODU5NF5BMl5BanBnXkFtZTgwNTU1NjM4NTE@._V1_UY1200_AL_.jpg'),
        # The Forest (mini series)
        ('tt6560040', 'https://m.media-amazon.com/images/M/MV5BMzY5NThkOWItN2I1OC00MzQ2LThlYjktMGExYTk2YzM1ZGRmXkEyXkFqcGdeQXVyODEyMzI2OTE@._V1_UY1200_AL_.jpg'),
        # Deadwood (series)
        ('tt0348914', 'https://m.media-amazon.com/images/M/MV5BNDJhMjUzMDYtNzc4MS00Nzk2LTkyMGQtN2M5NTczYTZmYmY5XkEyXkFqcGdeQXVyMzU3MTc5OTE@._V1_UY1200_AL_.jpg'),
        # Deadwood - S02E04 (episode)
        ('tt0556307', ''),
        (None, ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_poster_url(id, exp_poster_url, api, store_response):
    poster_url = await api.poster_url(id)
    assert poster_url == exp_poster_url


@pytest.mark.parametrize(
    argnames=('id', 'exp_rating'),
    argvalues=(
        ('tt0080455', 7.9),  # Blues Brothers (movie)
        ('tt0192802', 7.4),  # Wind in the Willows (TV movie)
        ('tt0471711', 7.6),  # Bender's Big Score (Video)
        ('tt0097270', 7.1),  # Elephant (TV movie)
        ('tt3472226', 8.0),  # Kung Fury (Short)
        ('tt6560040', 7.2),  # The Forest (mini series)
        ('tt0348914', 8.6),  # Deadwood (series)
        ('tt0556307', 8.2),  # Deadwood - S02E04 (episode)
        (None, None),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_rating(id, exp_rating, api, store_response):
    rating = await api.rating(id)
    assert rating == exp_rating


@pytest.mark.parametrize(
    argnames=('id', 'exp_runtimes'),
    argvalues=(
        ('tt0080455', {'default': 133, 'Extended': 148}),  # Blues Brothers (movie)
        ('tt0097270', {'default': 39}),  # Elephant (TV movie)
        ('tt0192802', {'default': 73}),  # Wind in the Willows (TV movie)
        ('tt0348914', {'default': 55}),  # Deadwood (series)
        ('tt0402123', {}),  # Deadwood Coach (movie, no runtime found)
        ('tt0409459', {'default': 162, "Director's Cut": 186, 'Ultimate Cut': 215}),  # Watchmen (movie)
        ('tt0556307', {'default': 53}),  # Deadwood - S02E04 (episode)
        ('tt3323824', {'default': 23}),  # Drifters (series) with removed "approx" runtime
        ('tt6560040', {'Entire Series': 313}),  # The Forest (mini series)
        (None, {}),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_runtimes(id, exp_runtimes, api, store_response):
    runtimes = await api.runtimes(id)
    assert runtimes == exp_runtimes


@pytest.mark.parametrize(
    argnames=('id', 'exp_summary'),
    argvalues=(
        ('tt0080455', ('Jake Blues rejoins with his brother Elwood after being released '
                       'from prison, but the duo has just days to reunite their old R&B band '
                       'and save the Catholic home where the two were raised, outrunning '
                       'the police as they tear through Chicago.')),  # Blues Brothers (movie)
        ('tt0192802', ('When three close friends Mole, Ratty and Badger '
                       'find out that the infamous Mr. Toad of Toad Hall '
                       'has been up to no good, they must find him and change '
                       'his ways for good.')),  # Wind in the Willows (TV movie)
        ('tt0471711', ('Planet Express sees a hostile takeover and Bender falls into the hands of '
                       'criminals where he is used to fulfill their schemes.')),  # Bender's Big Score (Video)
        ('tt0097270', ('A depiction of a series of violent killings in Northern Ireland '
                       'with no clue as to exactly who is responsible.')),  # Elephant (TV Short)
        ('tt3472226', ('In 1985, Kung Fury, the toughest martial artist cop in Miami, '
                       'goes back in time to kill the worst criminal of all time '
                       '- Kung Führer, a.k.a. Adolf Hitler.')),  # Kung Fury (Short)
        ('tt6560040', ('Sixteen-year-old Jennifer disappears one night from her village '
                       'in the Ardennes. Captain Gaspard Deker leads the investigation with '
                       'local cop Virginie Musso, who knew the girl well. They are helped '
                       'by Eve, a lonely and mysterious woman.')),  # The Forest (TV mini-series)
        ('tt0348914', ('A show set in the late 1800s, revolving around '
                       'the characters of Deadwood, South Dakota; a town '
                       'of deep corruption and crime.')),  # Deadwood (series)
        ('tt0556307', ('Doc contemplates a procedure that could cure Swearengen - or kill him. '
                       'Bullock attempts to settle into domesticity, while Sol gets a new '
                       "student bookkeeper - Trixie. Alma cuts ties with Sofia's tutor, "
                       'Miss Isringhausen. Joanie and Maddie argue over the business. The '
                       "County Commissioner's arrival spawns rumors about the camp's future "
                       'and legal ownership of the gold claims.')),  # Deadwood - S02E04 (episode)
        ('tt0014838', ''),  # The Deadwood Coach
        # Links in summary
        ('tt0200849', (
            # Mein liebster Feind
            ('In the 1950s, an adolescent Werner Herzog was transfixed by a '
             'film performance of the young Klaus Kinski. Years later, they would share '
             'an apartment where, in an unabated, forty-eight-hour fit of rage, Kinski '
             'completely destroyed the bathroom. From this chaos, a violent, love-hate, '
             'profoundly creative partnership was born. In 1972, Herzog cast Kinski in '
             'Aguirre, the Wrath of God (1972). Four more films would '
             'follow. In this personal documentary, Herzog traces the often violent ups and '
             'downs of their relationship, revisiting the various locations of their films '
             'and talking to the people they worked with.'),
            ('In the 1950s, an adolescent Werner Herzog was transfixed by a '
             'film performance of the young Klaus Kinski. Years later, they would share '
             'an apartment where, in an unabated, forty-eight-hour fit of rage, Kinski '
             'completely destroyed the bathroom. From this chaos, a violent, love-hate, '
             'profoundly creative partnership was born. In 1972, Herzog cast Kinski in '
             'Aguirre, der Zorn Gottes (1972). Four more films would '
             'follow. In this personal documentary, Herzog traces the often violent ups and '
             'downs of their relationship, revisiting the various locations of their films '
             'and talking to the people they worked with.'),
        )),
        (None, ''),
    ),
    ids=lambda value: str(value)[:30] or '<empty>',
)
@pytest.mark.asyncio
async def test_summary(id, exp_summary, api, store_response):
    summary = await api.summary(id)
    if isinstance(exp_summary, str):
        assert summary == exp_summary
    else:
        print('summary:', summary)
        print('expected:')
        for exp_sum in exp_summary:
            print(exp_sum)
        assert any(summary == exp_sum for exp_sum in exp_summary)


@pytest.mark.parametrize(
    argnames=('id', 'exp_title_english', 'exp_title_original'),
    argvalues=(
        ('tt0048918', '', '1984'),
        ('tt0078243', 'The 36th Chamber of Shaolin', 'Shao Lin san shi liu fang'),
        ('tt0080455', '', 'The Blues Brothers'),
        ('tt0082242', '', 'Dead & Buried'),
        ('tt0097138', '', 'Cyborg'),
        ('tt0102984', '', 'Stone Cold'),
        ('tt0104684', 'Hard Boiled', 'Lat sau san taam'),
        ('tt0188030', 'Butterfly', 'La lengua de las mariposas'),
        ('tt0280990', 'The Nest', 'Nid de guêpes'),
        ('tt0348914', '', 'Deadwood'),
        ('tt0396184', '', 'Pusher II'),
        ('tt1127715', '', 'Sin nombre'),
        ('tt1405737', 'Traffic Light', 'Ramzor'),
        ('tt1413492', '', '12 Strong'),
        ('tt1520211', '', 'The Walking Dead'),
        ('tt1526578', '', 'Ao, le dernier Néandertal'),
        ('tt1530509', '', 'The Human Centipede II (Full Sequence)'),
        ('tt1572491', 'The Last Circus', 'Balada triste de trompeta'),
        ('tt1866210', '', 'Wrecked'),  # No AKAs available
        ('tt2172934', '', '3 Days to Kill'),
        ('tt2209300', 'Aftermath', 'Poklosie'),
        ('tt2302755', '', 'Olympus Has Fallen'),
        ('tt3098998', '', 'Psychophonia'),
        ('tt3138376', '', 'Joy Ride 3: Road Kill'),
        ('tt3286052', "The Blackcoat's Daughter", 'February'),
        ('tt6560040', 'The Forest', 'La forêt'),
        ('tt6616260', 'Deadwind', 'Karppi'),
        ('tt7534328', '', 'Terror in the Woods'),
        ('tt9319668', '', '1899'),
        ('tt9616700', '', 'Adú'),
        (None, '', ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_title_english_original(id, exp_title_english, exp_title_original, api, store_response):
    title_original = await api.title_original(id)
    title_english = await api.title_english(id)
    print('original:', title_original, '-- expected:', exp_title_original)
    print('english:', title_english, '-- expected:', exp_title_english)
    assert title_original == exp_title_original
    assert title_english == exp_title_english
    assert await api.title_english(id, default_to_original=False) == exp_title_english
    if exp_title_english:
        assert await api.title_english(id, default_to_original=True) == exp_title_english
    else:
        assert await api.title_english(id, default_to_original=True) == exp_title_original


@pytest.mark.parametrize(
    argnames=('id', 'exp_type'),
    argvalues=(
        ('tt3286052', ReleaseType.movie),    # February AKA The Blackcoat's Daughter
        ('tt0080455', ReleaseType.movie),    # Blues Brothers (movie)
        ('tt0192802', ReleaseType.movie),    # Wind in the Willows (TV movie)
        ('tt0471711', ReleaseType.movie),    # Bender's Big Score (Video)
        ('tt0097270', ReleaseType.movie),    # Elephant (TV movie)
        ('tt3472226', ReleaseType.movie),    # Kung Fury (Short)
        ('tt0463986', ReleaseType.movie),    # The Feast (short)
        ('tt0048918', ReleaseType.movie),    # 1984 (movie)
        ('tt6560040', ReleaseType.season),   # The Forest (mini series)
        ('tt0348914', ReleaseType.season),   # Deadwood (series)
        ('tt0556307', ReleaseType.episode),  # Deadwood - S02E04 (episode)
        (None, ReleaseType.unknown),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_type(id, exp_type, api, store_response):
    assert await api.type(id) == exp_type


@pytest.mark.asyncio
async def test_url(api):
    assert await api.url('foo') == api._url_base + '/title/foo'
    assert await api.url(None) == ''


@pytest.mark.parametrize(
    argnames=('id', 'exp_year'),
    argvalues=(
        ('tt0080455', '1980'),  # Blues Brothers (movie)
        ('tt0192802', '1995'),  # Wind in the Willows (TV movie)
        ('tt0471711', '2007'),  # Bender's Big Score (Video)
        ('tt0097270', '1989'),  # Elephant (TV Short)
        ('tt3472226', '2015'),  # Kung Fury (Short)
        ('tt6560040', '2017'),  # The Forest (mini series)
        ('tt0348914', '2004'),  # Deadwood (series)
        ('tt0556307', '2005'),  # Deadwood - S02E04 (episode)
        ('tt0048918', '1956'),  # 1984 (movie)
        ('tt9319668', '2022'),  # 1899 (series)
        ('tt2917940', ''),      # Thale 2 (unreleased movie)
        (None, ''),
    ),
    ids=lambda value: str(value),
)
@pytest.mark.asyncio
async def test_year(id, exp_year, api, store_response):
    assert await api.year(id) == exp_year



import asyncio
import inspect
from unittest.mock import AsyncMock, Mock, PropertyMock, call

import pytest

from upsies.jobs import JobBase, QueueJobBase


class FooJob(QueueJobBase):
    name = 'foo'
    label = 'Foo'

    def initialize(self, foo=None, bar=None, enqueue=()):
        self.handled_inputs = []

    async def handle_input(self, value):
        self.handled_inputs.append(value)


@pytest.fixture
@pytest.mark.asyncio
async def qjob(tmp_path):
    qjob = FooJob(home_directory=tmp_path, cache_directory=tmp_path, ignore_cache=False)
    yield qjob
    if inspect.isawaitable(qjob._read_queue_task):
        qjob.finish()
        try:
            await qjob.wait()
        except BaseException:
            pass


def test_QueueJobBase_is_JobBase_subclass(qjob):
    assert isinstance(qjob, JobBase)


def test_enqueue_argument_is_empty(mocker, tmp_path):
    mocker.patch('upsies.jobs.base.QueueJobBase.enqueue')
    mocker.patch('upsies.jobs.base.QueueJobBase.finalize')
    qjob = FooJob(home_directory=tmp_path, cache_directory=tmp_path, enqueue=())
    assert qjob._enqueue_args == ()
    assert qjob.enqueue.call_args_list == []
    assert qjob.finalize.call_args_list == []

def test_enqueue_argument_is_not_empty(mocker, tmp_path):
    mocks = Mock()
    mocker.patch('upsies.jobs.base.QueueJobBase.enqueue', mocks.enqueue)
    mocker.patch('upsies.jobs.base.QueueJobBase.finish', mocks.finish)
    qjob = FooJob(home_directory=tmp_path, cache_directory=tmp_path, enqueue=(1, 2, 3))
    assert qjob._enqueue_args == (1, 2, 3)
    assert mocks.mock_calls == []


def test_execute_with_enqueue_argument(mocker, tmp_path):
    mocks = Mock()
    mocker.patch('upsies.jobs.base.QueueJobBase.enqueue', mocks.enqueue)
    mocker.patch('upsies.jobs.base.QueueJobBase.finalize', mocks.finalize)
    mocker.patch('upsies.jobs.base.QueueJobBase._read_queue', Mock())
    mocker.patch('asyncio.ensure_future')
    qjob = FooJob(home_directory=tmp_path, cache_directory=tmp_path, enqueue=(1, 2, 3))
    assert qjob._read_queue_task is None
    qjob.execute()
    assert qjob._read_queue.call_args_list == [call()]
    assert asyncio.ensure_future.call_args_list == [call(qjob._read_queue.return_value)]
    assert qjob._read_queue_task == asyncio.ensure_future.return_value
    assert mocks.mock_calls == [
        call.enqueue(1),
        call.enqueue(2),
        call.enqueue(3),
        call.finalize(),
    ]

def test_execute_without_enqueue_argument(mocker, tmp_path):
    mocks = Mock()
    mocker.patch('upsies.jobs.base.QueueJobBase.enqueue', mocks.enqueue)
    mocker.patch('upsies.jobs.base.QueueJobBase.finalize', mocks.finalize)
    mocker.patch('upsies.jobs.base.QueueJobBase._read_queue', Mock())
    mocker.patch('asyncio.ensure_future')
    qjob = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    assert qjob._read_queue_task is None
    qjob.execute()
    assert qjob._read_queue.call_args_list == [call()]
    assert asyncio.ensure_future.call_args_list == [call(qjob._read_queue.return_value)]
    assert qjob._read_queue_task == asyncio.ensure_future.return_value
    assert mocks.mock_calls == []


@pytest.mark.asyncio
async def test_read_queue_reads_queue(qjob, mocker):
    mocker.patch.object(qjob, 'handle_input', AsyncMock())
    qjob._queue.put_nowait('a')
    qjob._queue.put_nowait('b')
    qjob._queue.put_nowait(None)
    await qjob._read_queue()
    assert qjob._queue.empty()
    assert qjob.handle_input.call_args_list == [
        call('a'),
        call('b'),
    ]

@pytest.mark.asyncio
async def test_read_queue_breaks_if_job_is_finished(mocker, tmp_path, event_loop):
    mocker.patch('upsies.jobs.base.QueueJobBase.is_finished', PropertyMock(
        side_effect=(False, True, True, True),
    ))
    qjob = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    mocker.patch.object(qjob, 'handle_input', AsyncMock())

    qjob.start()

    def finish():
        qjob.finish()
        qjob._queue.put_nowait('x')

    event_loop.call_later(1, finish)

    qjob._queue.put_nowait('a')
    qjob._queue.put_nowait('b')
    qjob._queue.put_nowait('c')

    await qjob._read_queue()

    assert qjob.is_finished
    assert not qjob._queue.empty()
    assert qjob.handle_input.call_args_list == []

@pytest.mark.asyncio
async def test_read_queue_breaks_if_it_reads_None(mocker, tmp_path):
    mocker.patch('upsies.jobs.base.QueueJobBase.is_finished', PropertyMock(
        side_effect=(False, True, True, True),
    ))
    qjob = FooJob(home_directory=tmp_path, cache_directory=tmp_path)
    qjob.start()
    mocker.patch.object(qjob, 'handle_input', AsyncMock())
    qjob._queue.put_nowait('a')
    qjob._queue.put_nowait('b')
    qjob._queue.put_nowait('c')
    qjob._queue.put_nowait(None)
    await qjob._read_queue()
    assert qjob.is_finished
    assert qjob.handle_input.call_args_list == []

@pytest.mark.asyncio
async def test_read_queue_handles_CancelledError_from_handle_input(qjob, mocker):
    mocker.patch.object(qjob, 'handle_input', AsyncMock(
        side_effect=asyncio.CancelledError(),
    ))
    qjob.start()
    qjob._queue.put_nowait('a')
    await qjob._read_queue()
    assert qjob._queue.empty()
    assert qjob.handle_input.call_args_list == [call('a')]
    await qjob.wait()
    assert qjob.is_finished
    assert qjob.raised is None

@pytest.mark.asyncio
async def test_read_queue_handles_exception_from_handle_input(qjob, mocker):
    mocker.patch.object(qjob, 'handle_input', AsyncMock(
        side_effect=TypeError('argh'),
    ))
    qjob.start()
    qjob._queue.put_nowait('a')
    await qjob._read_queue()
    assert qjob._queue.empty()
    assert qjob.handle_input.call_args_list == [call('a')]
    with pytest.raises(TypeError, match='^argh$'):
        await qjob.wait()


@pytest.mark.asyncio
async def test_enqueue(qjob):
    assert qjob._queue.empty()
    qjob.enqueue('a')
    qjob.enqueue('b')
    qjob.enqueue('c')
    queued_values = []
    while not qjob._queue.empty():
        queued_values.append(qjob._queue.get_nowait())
    assert queued_values == ['a', 'b', 'c']


@pytest.mark.asyncio
async def test_finalize(qjob):
    assert qjob._queue.empty()
    qjob.finalize()
    queued_values = []
    while not qjob._queue.empty():
        queued_values.append(qjob._queue.get_nowait())
    assert queued_values == [None]


@pytest.mark.asyncio
async def test_finish_without_running_read_queue_task(qjob):
    qjob.start()
    assert not qjob.is_finished
    qjob.finish()
    assert qjob.is_finished

@pytest.mark.asyncio
async def test_finish_with_running_read_queue_task(qjob):
    qjob.start()
    assert not qjob.is_finished
    qjob.finish()
    assert qjob.is_finished


@pytest.mark.asyncio
async def test_wait_without_running_read_queue_task(qjob, event_loop):
    assert qjob._read_queue_task is None
    assert not qjob.is_finished
    event_loop.call_soon(qjob.start)
    event_loop.call_soon(qjob.finish)
    await qjob.wait()
    assert qjob.is_finished
    assert qjob.errors == ()

@pytest.mark.asyncio
async def test_wait_awaits_read_queue_task(qjob, event_loop):
    qjob.start()
    assert qjob._read_queue_task is not None
    assert not qjob.is_finished
    event_loop.call_soon(qjob.finalize)
    await qjob.wait()
    assert qjob.is_finished
    assert qjob.errors == ()
    assert qjob._read_queue_task.done()

@pytest.mark.asyncio
async def test_wait_handles_CancelledError_from_read_queue_task(qjob, event_loop):
    qjob.start()
    assert qjob._read_queue_task is not None
    for i in range(1000):
        qjob.enqueue(i)
    assert not qjob.is_finished
    event_loop.call_soon(qjob.finish)
    await qjob.wait()
    assert qjob.is_finished
    assert qjob.exit_code == 1
    assert qjob._read_queue_task.done()



import asyncio
import re
from unittest.mock import AsyncMock, Mock, call

import pytest

from upsies import errors
from upsies.utils.predbs import SceneCheckResult, multi


def test_with_custom_predbs(mocker):
    predb_mock = mocker.patch('upsies.utils.predbs.predb')
    predbs = (Mock(), Mock(), Mock())
    multipredb = multi.MultiPredbApi(predbs)
    assert multipredb._predbs == predbs
    assert predb_mock.call_args_list == []

def test_with_default_predbs(mocker):
    exp_predbs = [
        Mock(name=name)
        for name in multi.MultiPredbApi.DEFAULT_PREDB_NAMES
    ]
    predb_mock = mocker.patch('upsies.utils.predbs.predb', side_effect=exp_predbs)
    multipredb = multi.MultiPredbApi()
    assert multipredb._predbs == exp_predbs
    assert predb_mock.call_args_list == [
        call('srrdb'),
        # call('corruptnet'),
        # call('predbovh'),
        # call('predbclub'),
        call('predbde'),
    ]


@pytest.fixture
def multipredb():
    predbs = [Mock(), Mock(), Mock()]
    predbs[0].configure_mock(name='foo')
    predbs[1].configure_mock(name='bar')
    predbs[2].configure_mock(name='baz')
    return multi.MultiPredbApi(predbs=predbs)


def test_predbs(multipredb, mocker):
    mocker.patch.object(multipredb, '_predbs', ['mock predb 1', 'mock predb 2'])
    assert multipredb.predbs == ('mock predb 1', 'mock predb 2')


@pytest.mark.asyncio
async def test_call(multipredb, mocker):
    predb = Mock(my_method=AsyncMock(return_value='my response'))

    return_value = await multipredb._call(predb, 'my_method', 123, this='that')
    assert return_value == 'my response'
    assert predb.my_method.call_args_list == [call(123, this='that')]


@pytest.mark.parametrize(
    argnames='exceptions, method_name, exp_exception',
    argvalues=(
        (
            [],
            'my_method',
            NotImplementedError('foo|bar|baz.my_method'),
        ),
        (
            [errors.RequestError('only error')],
            'my_method',
            errors.RequestError('only error'),
        ),
        (
            [errors.RequestError('error one'), errors.RequestError('error two')],
            'my_method',
            errors.RequestError('All queries failed: error one, error two'),
        ),
    ),
    ids=lambda v: repr(v),
)
def test_raise(exceptions, method_name, exp_exception, multipredb, mocker):
    with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
        multipredb._raise(exceptions, method_name)


class MockPredb(str):
    def __init__(self, name):
        self.name = name

@pytest.mark.parametrize(
    argnames='method_name, args, kwargs, predbs, call_side_effects, exp_result, exp_mock_calls',
    argvalues=(
        pytest.param(
            'mymethod', ('foo',), {'this': 'that'},
            [MockPredb('predb1'), MockPredb('predb2'), MockPredb('predb3'), MockPredb('predb4')],
            [
                NotImplementedError('this is not implemented'),
                errors.RequestError('Service is down'),
                'here is your response',
                NotImplementedError('that is not implemented'),
            ],
            'here is your response',
            [
                call('predb1', 'mymethod', 'foo', this='that'),
                call('predb2', 'mymethod', 'foo', this='that'),
                call('predb3', 'mymethod', 'foo', this='that'),
            ],
            id='First good response is returned',
        ),

        pytest.param(
            'mymethod', ('foo',), {'bar': 'baz'},
            [MockPredb('predb1'), MockPredb('predb2'), MockPredb('predb3'), MockPredb('predb4')],
            [
                NotImplementedError('This is not implemented'),
                errors.RequestError('Service is down'),
                NotImplementedError('This is also not implemented'),
                NotImplementedError('That is not implemented'),
            ],
            errors.RequestError(
                'All queries failed: '
                'This is not implemented, '
                'Service is down, '
                'This is also not implemented, '
                'That is not implemented'
            ),
            [
                call('predb1', 'mymethod', 'foo', bar='baz'),
                call('predb2', 'mymethod', 'foo', bar='baz'),
                call('predb3', 'mymethod', 'foo', bar='baz'),
                call('predb4', 'mymethod', 'foo', bar='baz'),
            ],
            id='All raise',
        ),
    ),
)
@pytest.mark.asyncio
async def test_for_each_predb(method_name, args, kwargs, predbs, call_side_effects, exp_result, exp_mock_calls, multipredb, mocker):
    mocker.patch.object(multipredb, '_call', side_effect=call_side_effects)
    mocker.patch.object(multipredb, '_predbs', predbs)

    if isinstance(exp_result, Exception):
        with pytest.raises(type(exp_result), match=rf'^{re.escape(str(exp_result))}$'):
            await multipredb._for_each_predb(method_name, *args, **kwargs)
    else:
        return_value = await multipredb._for_each_predb(method_name, *args, **kwargs)
        assert return_value == exp_result

    assert multipredb._call.call_args_list == exp_mock_calls


@pytest.mark.asyncio
async def test_for_all_predbs(multipredb, mocker):
    predbs = [MockPredb(name) for name in 'abcde']
    mocker.patch.object(multipredb, '_predbs', predbs)

    async def call_side_effect(predb, method_name, *args, **kwargs):
        if predb.name == 'a':
            await asyncio.sleep(0.3)
            raise errors.RequestError('4: a')
        elif predb.name == 'b':
            await asyncio.sleep(0.0)
            raise NotImplementedError('1: b')
        elif predb.name == 'c':
            await asyncio.sleep(0.2)
            return '3: c: result'
        elif predb.name == 'd':
            await asyncio.sleep(0.4)
            return '5: d: result'
        elif predb.name == 'e':
            await asyncio.sleep(0.1)
            return '2: e: result'
        else:
            raise RuntimeError(f'Unexpected predb: {predb.name}')

    mocker.patch.object(multipredb, '_call', side_effect=call_side_effect)

    method_name = 'mymethod'
    args = ('foo',)
    kwargs = {'bar': 'baz'}

    exp_results = [
        NotImplementedError('1: b'),
        '2: e: result',
        '3: c: result',
        errors.RequestError('4: a'),
        '5: d: result',
    ]

    async for result in multipredb._for_all_predbs(method_name, *args, **kwargs):
        exp_result = exp_results.pop(0)
        assert str(result) == str(exp_result)

    assert multipredb._call.call_args_list == [
        call(predbs[0], method_name, *args, **kwargs),
        call(predbs[1], method_name, *args, **kwargs),
        call(predbs[2], method_name, *args, **kwargs),
        call(predbs[3], method_name, *args, **kwargs),
        call(predbs[4], method_name, *args, **kwargs),
    ]


@pytest.fixture
def multipredb_for_verify_release(multipredb, mocker):
    mocks = Mock()
    mocks.attach_mock(mocker.patch.object(multipredb, 'search'), 'search')
    mocks.attach_mock(mocker.patch.object(multipredb, '_verify_release'), '_verify_release')
    mocks.attach_mock(mocker.patch.object(multipredb, '_verify_release_per_file'), '_verify_release_per_file')
    return multipredb, mocks

@pytest.mark.asyncio
async def test_verify_release_gets_release_name(multipredb_for_verify_release):
    multipredb, mocks = multipredb_for_verify_release
    content_path = 'path/to/Mock.Release'
    release_name = 'Mock.Release'

    return_value = await multipredb.verify_release(content_path, release_name)
    assert return_value is mocks._verify_release.return_value

    assert mocks.mock_calls == [
        call._verify_release(content_path, release_name),
    ]

@pytest.mark.asyncio
async def test_verify_release_gets_no_release_name_and_finds_no_matching_scene_release(multipredb_for_verify_release):
    multipredb, mocks = multipredb_for_verify_release
    content_path = 'path/to/Mock.Release'
    mocks.search.return_value = ()

    return_value = await multipredb.verify_release(content_path)
    assert return_value == (SceneCheckResult.false, ())

    assert mocks.mock_calls == [
        call.search(content_path),
    ]

@pytest.mark.asyncio
async def test_verify_release_finds_valid_scene_release(multipredb_for_verify_release):
    multipredb, mocks = multipredb_for_verify_release
    content_path = 'path/to/Mock.Release'
    verify_release_results = {
        'The.Foo.2000.x264-AAA': (SceneCheckResult.false, ()),
        'The.Foo.2000.x264-BBB': (SceneCheckResult.unknown, ()),
        'The.Foo.2000.x264-CCC': (SceneCheckResult.true, (errors.SceneRenamedError(original_name='foo', existing_name='bar'),)),
        'The.Foo.2000.x264-DDD': (SceneCheckResult.true, ()),
        'The.Foo.2000.x264-EEE': (SceneCheckResult.true, (errors.SceneFileSizeError('foo', original_size=123, existing_size=456),)),
    }
    mocks.search.return_value = tuple(verify_release_results.keys())
    mocks._verify_release.side_effect = tuple(verify_release_results.values())

    return_value = await multipredb.verify_release(content_path)
    assert return_value == (SceneCheckResult.true, ())

    assert mocks.mock_calls == [
        call.search(content_path),
        call._verify_release(content_path, 'The.Foo.2000.x264-AAA'),
        call._verify_release(content_path, 'The.Foo.2000.x264-BBB'),
        call._verify_release(content_path, 'The.Foo.2000.x264-CCC'),
        call._verify_release(content_path, 'The.Foo.2000.x264-DDD'),
    ]

@pytest.mark.asyncio
async def test_verify_release_finds_only_nonscene_releases(multipredb_for_verify_release):
    multipredb, mocks = multipredb_for_verify_release
    content_path = 'path/to/Mock.Release'
    verify_release_results = {
        'The.Foo.2000.x264-AAA': (SceneCheckResult.false, ()),
        'The.Foo.2000.x264-BBB': (SceneCheckResult.unknown, ()),
    }
    mocks.search.return_value = tuple(verify_release_results.keys())
    mocks._verify_release.side_effect = tuple(verify_release_results.values())

    return_value = await multipredb.verify_release(content_path)
    assert return_value is mocks._verify_release_per_file.return_value

    assert mocks.mock_calls == [
        call.search(content_path),
        call._verify_release(content_path, 'The.Foo.2000.x264-AAA'),
        call._verify_release(content_path, 'The.Foo.2000.x264-BBB'),
        call._verify_release_per_file(content_path),
    ]

@pytest.mark.asyncio
async def test_verify_release_finds_only_scene_releases_with_exceptions(multipredb_for_verify_release):
    multipredb, mocks = multipredb_for_verify_release
    content_path = 'path/to/Mock.Release'
    verify_release_results = {
        'The.Foo.2000.x264-AAA': (SceneCheckResult.true, (errors.SceneRenamedError(original_name='foo', existing_name='bar'),)),
        'The.Foo.2000.x264-BBB': (SceneCheckResult.unknown, ()),
        'The.Foo.2000.x264-CCC': (SceneCheckResult.true, (errors.SceneFileSizeError('foo', original_size=123, existing_size=456),)),
    }
    mocks.search.return_value = tuple(verify_release_results.keys())
    mocks._verify_release.side_effect = tuple(verify_release_results.values())

    return_value = await multipredb.verify_release(content_path)
    assert return_value is mocks._verify_release_per_file.return_value

    assert mocks.mock_calls == [
        call.search(content_path),
        call._verify_release(content_path, 'The.Foo.2000.x264-AAA'),
        call._verify_release(content_path, 'The.Foo.2000.x264-BBB'),
        call._verify_release(content_path, 'The.Foo.2000.x264-CCC'),
        call._verify_release_per_file(content_path),
    ]


@pytest.fixture
def multipredb_for__verify_release(multipredb, mocker):
    mocks = Mock()
    mocks.attach_mock(
        mocker.patch.object(multipredb, 'is_scene_release'),
        'is_scene_release',
    )
    mocks.attach_mock(
        mocker.patch.object(multipredb, 'verify_release_name'),
        'verify_release_name',
    )
    mocks.attach_mock(
        mocker.patch.object(multipredb, 'verify_release_files'),
        'verify_release_files',
    )
    return multipredb, mocks

@pytest.mark.asyncio
async def test__verify_release_gets_nonscene_release_name(multipredb_for__verify_release):
    multipredb, mocks = multipredb_for__verify_release
    mocks.is_scene_release.return_value = SceneCheckResult.false

    is_scene, exceptions = await multipredb._verify_release('mock/path', 'Mock.Release')
    assert is_scene is SceneCheckResult.false
    assert exceptions == ()
    assert mocks.mock_calls == [
        call.is_scene_release('Mock.Release'),
    ]

@pytest.mark.asyncio
async def test__verify_release_gets_release_with_unknown_scene_status(multipredb_for__verify_release):
    multipredb, mocks = multipredb_for__verify_release
    mocks.is_scene_release.return_value = SceneCheckResult.unknown

    is_scene, exceptions = await multipredb._verify_release('mock/path', 'Mock.Release')
    assert is_scene is SceneCheckResult.unknown
    assert exceptions == ()
    assert mocks.mock_calls == [
        call.is_scene_release('Mock.Release'),
    ]

@pytest.mark.asyncio
async def test__verify_release_gets_wrong_release_name(multipredb_for__verify_release):
    multipredb, mocks = multipredb_for__verify_release
    mocks.is_scene_release.return_value = SceneCheckResult.true
    mocks.verify_release_name.side_effect = errors.SceneRenamedError(original_name='foo', existing_name='bar')
    mocks.verify_release_files.return_value = ()

    is_scene, exceptions = await multipredb._verify_release('mock/path', 'Mock.Release')
    assert is_scene is SceneCheckResult.true
    assert exceptions == (errors.SceneRenamedError(original_name='foo', existing_name='bar'),)
    assert mocks.mock_calls == [
        call.is_scene_release('Mock.Release'),
        call.verify_release_name('mock/path', 'Mock.Release'),
        call.verify_release_files('mock/path', 'Mock.Release'),
    ]

@pytest.mark.asyncio
async def test__verify_release_gets_wrong_release_files(multipredb_for__verify_release):
    multipredb, mocks = multipredb_for__verify_release
    mocks.is_scene_release.return_value = SceneCheckResult.true
    mocks.verify_release_files.return_value = (
        errors.SceneFileSizeError('foo', original_size=123, existing_size=456),
        errors.SceneFileSizeError('bar', original_size=100, existing_size=200),
    )

    is_scene, exceptions = await multipredb._verify_release('mock/path', 'Mock.Release')
    assert is_scene is SceneCheckResult.true
    assert exceptions == (
        errors.SceneFileSizeError('foo', original_size=123, existing_size=456),
        errors.SceneFileSizeError('bar', original_size=100, existing_size=200),
    )
    assert mocks.mock_calls == [
        call.is_scene_release('Mock.Release'),
        call.verify_release_name('mock/path', 'Mock.Release'),
        call.verify_release_files('mock/path', 'Mock.Release'),
    ]

@pytest.mark.asyncio
async def test__verify_release_gets_wrong_release_name_and_wrong_files(multipredb_for__verify_release):
    multipredb, mocks = multipredb_for__verify_release
    mocks.is_scene_release.return_value = SceneCheckResult.true
    mocks.verify_release_name.side_effect = errors.SceneRenamedError(original_name='foo', existing_name='bar')
    mocks.verify_release_files.return_value = (
        errors.SceneFileSizeError('foo', original_size=123, existing_size=456),
        errors.SceneFileSizeError('bar', original_size=100, existing_size=200),
    )

    is_scene, exceptions = await multipredb._verify_release('mock/path', 'Mock.Release')
    assert is_scene is SceneCheckResult.true
    assert exceptions == (
        errors.SceneRenamedError(original_name='foo', existing_name='bar'),
        errors.SceneFileSizeError('foo', original_size=123, existing_size=456),
        errors.SceneFileSizeError('bar', original_size=100, existing_size=200),
    )
    assert mocks.mock_calls == [
        call.is_scene_release('Mock.Release'),
        call.verify_release_name('mock/path', 'Mock.Release'),
        call.verify_release_files('mock/path', 'Mock.Release'),
    ]


@pytest.mark.parametrize(
    argnames='content_path, mock_data, exp_mock_calls, exp_is_scene_release, exp_exceptions',
    argvalues=(
        pytest.param(
            'path/to/content',
            {
                'foo.mkv': (
                    {'release_name': 'Foo-AAA', 'is_scene_release': SceneCheckResult.false, 'exceptions': ()},
                    {'release_name': 'Foo-BBB', 'is_scene_release': SceneCheckResult.true, 'exceptions': ()},
                ),
                'bar.mp4': (
                    {'release_name': 'Bar-AAA', 'is_scene_release': SceneCheckResult.true, 'exceptions': ()},
                    {'release_name': 'Bar-BBB', 'is_scene_release': SceneCheckResult.true,
                     'exceptions': (errors.SceneError('never found'),)},
                ),
            },
            [
                call.search('foo.mkv'),
                call._verify_release('foo.mkv', 'Foo-AAA'),
                call._verify_release('foo.mkv', 'Foo-BBB'),
                call.search('bar.mp4'),
                call._verify_release('bar.mp4', 'Bar-AAA'),
            ],
            SceneCheckResult.true,
            (),
            id='Verification is stopped after first match is found',
        ),

        pytest.param(
            'path/to/content',
            {
                'foo.mkv': (
                    {'release_name': 'Foo-AAA', 'is_scene_release': SceneCheckResult.false, 'exceptions': ()},
                    {'release_name': 'Foo-BBB', 'is_scene_release': SceneCheckResult.unknown, 'exceptions': ()},
                ),
                'bar.mp4': (
                    {'release_name': 'Bar-AAA', 'is_scene_release': SceneCheckResult.true, 'exceptions': ()},
                    {'release_name': 'Bar-BBB', 'is_scene_release': SceneCheckResult.true,
                     'exceptions': (errors.SceneError('never found'),)},
                ),
            },
            [
                call.search('foo.mkv'),
                call._verify_release('foo.mkv', 'Foo-AAA'),
                call._verify_release('foo.mkv', 'Foo-BBB'),
                call.search('bar.mp4'),
                call._verify_release('bar.mp4', 'Bar-AAA'),
            ],
            SceneCheckResult.unknown,
            (),
            id='One scene check is unknown',
        ),

        pytest.param(
            'path/to/content',
            {
                'foo.mkv': (
                    {'release_name': 'Foo-AAA', 'is_scene_release': SceneCheckResult.false, 'exceptions': ()},
                    {'release_name': 'Foo-BBB', 'is_scene_release': SceneCheckResult.false, 'exceptions': ()},
                ),
                'bar.mp4': (
                    {'release_name': 'Bar-AAA', 'is_scene_release': SceneCheckResult.false, 'exceptions': ()},
                    {'release_name': 'Bar-BBB', 'is_scene_release': SceneCheckResult.false, 'exceptions': ()},
                ),
            },
            [
                call.search('foo.mkv'),
                call._verify_release('foo.mkv', 'Foo-AAA'),
                call._verify_release('foo.mkv', 'Foo-BBB'),
                call.search('bar.mp4'),
                call._verify_release('bar.mp4', 'Bar-AAA'),
                call._verify_release('bar.mp4', 'Bar-BBB'),
            ],
            SceneCheckResult.false,
            (),
            id='All scene checks are false',
        ),

        pytest.param(
            'path/to/content',
            {
                'foo.mkv': (
                    {'release_name': 'Foo-AAA', 'is_scene_release': SceneCheckResult.true, 'exceptions': (errors.SceneError('foo!'),)},
                    {'release_name': 'Foo-BBB', 'is_scene_release': SceneCheckResult.true, 'exceptions': ()},
                ),
                'bar.mp4': (
                    {'release_name': 'Bar-AAA', 'is_scene_release': SceneCheckResult.true, 'exceptions': ()},
                    {'release_name': 'Bar-BBB', 'is_scene_release': SceneCheckResult.true, 'exceptions': ()},
                ),
            },
            [
                call.search('foo.mkv'),
                call._verify_release('foo.mkv', 'Foo-AAA'),
                call._verify_release('foo.mkv', 'Foo-BBB'),
                call.search('bar.mp4'),
                call._verify_release('bar.mp4', 'Bar-AAA'),
            ],
            SceneCheckResult.true,
            (errors.SceneError('foo!'),),
            id='One scene check returns exceptions',
        ),

        pytest.param(
            'path/to/content',
            {
                'foo.mkv': (
                    {'release_name': 'Foo-AAA', 'is_scene_release': SceneCheckResult.true, 'exceptions': (errors.SceneError('foo!'),)},
                    {'release_name': 'Foo-BBB', 'is_scene_release': SceneCheckResult.true, 'exceptions': ()},
                ),
                'bar.mp4': (
                    {'release_name': 'Bar-AAA', 'is_scene_release': SceneCheckResult.false, 'exceptions': ()},
                    {'release_name': 'Bar-BBB', 'is_scene_release': SceneCheckResult.true, 'exceptions': (errors.SceneError('bar!'),)},
                ),
            },
            [
                call.search('foo.mkv'),
                call._verify_release('foo.mkv', 'Foo-AAA'),
                call._verify_release('foo.mkv', 'Foo-BBB'),
                call.search('bar.mp4'),
                call._verify_release('bar.mp4', 'Bar-AAA'),
                call._verify_release('bar.mp4', 'Bar-BBB'),
            ],
            SceneCheckResult.true,
            (errors.SceneError('foo!'), errors.SceneError('bar!'), ),
            id='Multiple scene checks return exceptions',
        ),
    ),
)
@pytest.mark.asyncio
async def test_verify_release_per_file(content_path, mock_data, exp_mock_calls, exp_is_scene_release, exp_exceptions, multipredb, mocker):
    mocks = Mock()

    mocker.patch('upsies.utils.fs.file_list', Mock(
        return_value=mock_data.keys(),
    ))
    mocks.attach_mock(
        mocker.patch.object(multipredb, 'search', AsyncMock(side_effect=(
            tuple(data['release_name'] for data in datas)
            for datas in mock_data.values()
        ))),
        'search',
    )
    mocks.attach_mock(
        mocker.patch.object(multipredb, '_verify_release', AsyncMock(side_effect=tuple(
            (data['is_scene_release'], data['exceptions'])
            for datas in mock_data.values()
            for data in datas
        ))),
        '_verify_release',
    )

    is_scene_release, exceptions = await multipredb._verify_release_per_file(content_path)
    assert is_scene_release == exp_is_scene_release
    assert exceptions == exp_exceptions
    assert mocks.mock_calls == exp_mock_calls



import base64
import re

import pytest

from upsies import utils
from upsies.trackers.ptp import PtpTrackerConfig


@pytest.fixture
def tracker_config():
    return PtpTrackerConfig()


def test_PtpTrackerConfig_defaults(tracker_config):
    assert set(tracker_config) == {
        'base_url',
        'username',
        'password',
        'announce_url',
        'source',
        'randomize_infohash',
        'image_host',
        'screenshots_from_movie',
        'screenshots_from_episode',
        'exclude',

        # Inherited from TrackerConfigBase
        'add_to',
        'copy_to',
    }


def test_PtpTrackerConfig_defaults_base_url(tracker_config):
    assert tracker_config['base_url'] == base64.b64decode('aHR0cHM6Ly9wYXNzdGhlcG9wY29ybi5tZQ==').decode('ascii')


def test_PtpTrackerConfig_defaults_username(tracker_config):
    assert tracker_config['username'] == ''


def test_PtpTrackerConfig_defaults_password(tracker_config):
    assert tracker_config['password'] == ''


def test_PtpTrackerConfig_defaults_announce_url(tracker_config):
    assert tracker_config['announce_url'] == ''
    assert tracker_config['announce_url'].description == (
        'Your personal announce URL.'
    )


def test_PtpTrackerConfig_defaults_source(tracker_config):
    assert tracker_config['source'] == 'PTP'


def test_PtpTrackerConfig_defaults_image_host(tracker_config, assert_config_choice):
    assert_config_choice(
        choice=tracker_config['image_host'],
        value='ptpimg',
        options=utils.imghosts.imghost_names(),
    )


def test_PtpTrackerConfig_defaults_screenshots_from_movie(tracker_config, assert_config_number):
    assert_config_number(
        number=tracker_config['screenshots_from_movie'],
        value=3,
        min=3,
        max=10,
        description='How many screenshots to make for single-video uploads.',
    )


def test_PtpTrackerConfig_defaults_screenshots_from_episode(tracker_config, assert_config_number):
    assert_config_number(
        number=tracker_config['screenshots_from_episode'],
        value=2,
        min=2,
        max=10,
        description='How many screenshots to make per video for multi-video uploads.',
    )


def test_PtpTrackerConfig_defaults_exclude(tracker_config):
    assert tracker_config['exclude'] == (
        re.compile(r'\.(?i:nfo|txt|jpg|jpeg|png|sfv|md5)$'),
        re.compile(r'/(?i:sample|extra|bonus|feature)'),
        re.compile(r'(?i:sample\.[a-z]+)$'),
    )


def test_MtvTrackerConfig_arguments(tracker_config):
    exp_argument_definitions = {
        'submit': {
            ('--not-main-movie', '--nmm'),
            ('--personal-rip', '--pr'),
            ('--screenshots', '--ss'),
            ('--upload-token', '--ut'),
            ('--only-description', '--od'),
        },
    }
    assert set(tracker_config.argument_definitions) == set(exp_argument_definitions)
    for command in exp_argument_definitions:
        assert set(tracker_config.argument_definitions[command]) == exp_argument_definitions[command]


def test_PtpTrackerConfig_argument_definitions_submit_not_main_movie(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--not-main-movie', '--nmm')] == {
        'help': 'Upload ONLY contains extras, Rifftrax, Workprints',
        'action': 'store_true',
    }


def test_PtpTrackerConfig_argument_definitions_submit_personal_rip(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--personal-rip', '--pr')] == {
        'help': 'Tag as your own encode',
        'action': 'store_true',
    }


def test_PtpTrackerConfig_argument_definitions_submit_screenshots(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--screenshots', '--ss')] == {
        'help': 'How many screenshots to make per video file',
        'type': utils.argtypes.number_of_screenshots(min=1, max=10),
    }


def test_PtpTrackerConfig_argument_definitions_submit_upload_token(tracker_config):
    assert tracker_config.argument_definitions['submit'][('--upload-token', '--ut')] == {
        'help': 'Upload token from staff',
    }



import os
from unittest.mock import Mock, patch

import pytest


# Platform independent fs.filename_sanitization() mocking fixture from
# tests/conftest.py
@pytest.fixture(scope='module', autouse=True)
def strict_filename_sanitization(strict_filename_sanitization):
    pass


@pytest.fixture(scope='module', autouse=True)
def disable_http_requests(pytestconfig, module_mocker):
    if not pytestconfig.getoption('--allow-requests', None):
        exc = RuntimeError('HTTP requests are disabled; use --allow-requests')

        # We can't patch utils.http._request() because we want it to return
        # cached requests. utils.http._request() only uses
        # httpx.AsyncClient.send() so we can patch that.
        module_mocker.patch('httpx.AsyncClient.send', Mock(side_effect=exc))


# When HTTP requests are allowed, store responses tests/data/webdbs.
# See tests/conftest.py for the data_dir fixture.
@pytest.fixture(scope='session')
def store_response(data_dir):
    cache_dir = os.path.join(data_dir, 'predbs')
    if not os.path.exists(cache_dir):
        os.mkdir(cache_dir)
    with patch('upsies.constants.DEFAULT_CACHE_DIRECTORY', cache_dir):
        yield



import re
from unittest.mock import Mock, call

import pytest

from upsies.utils import subtitle


@pytest.mark.parametrize('forced, exp_forced', (
    (True, True),
    (False, False),
    ('', False),
    (1, True),
))
@pytest.mark.parametrize('format, exp_format', (
    ('SRT', 'SRT'),
    ('', ''),
    (None, ''),
))
@pytest.mark.parametrize(
    argnames='language, exp_language, exp_attributes, exp_exception',
    argvalues=(
        ('foo-mx-Latn', None, None, ValueError("Invalid language: 'foo-mx-Latn'")),

        ('pt', 'pt', {'language': 'pt', 'territory': ''}, None),
        ('PT', 'pt', {'language': 'pt', 'territory': ''}, None),
        ('por', 'pt', {'language': 'pt', 'territory': ''}, None),

        ('Pt-bR', 'pt-BR', {'language': 'pt', 'territory': 'BR'}, None),
        ('spa-419', 'es-419', {'language': 'es', 'territory': '419'}, None),
        ('zh-HANS', 'zh-Hans', {'language': 'zh', 'territory': ''}, None),
        ('gsw-u-sd-chzh', 'gsw-u-sd-chzh', {'language': 'gsw', 'territory': ''}, None),
    ),
)
def test_Subtitle(language, exp_language, exp_attributes, exp_exception,
                  format, exp_format,
                  forced, exp_forced):
    if exp_exception:
        with pytest.raises(type(exp_exception), match=rf'^{re.escape(str(exp_exception))}$'):
            subtitle.Subtitle(language=language, forced=forced, format=format)
    else:
        sub = subtitle.Subtitle(language=language, forced=forced, format=format)
        assert sub.language == exp_attributes['language']
        assert sub.territory == exp_attributes['territory']
        assert sub.forced == exp_forced
        assert sub.format == exp_format

        assert repr(sub) == (
            'Subtitle('
            + repr(exp_language)
            + f', forced={exp_forced!r}'
            + f', format={exp_format!r}'
            + ')'
        )


@pytest.mark.parametrize(
    argnames='a, b, exp_equal',
    argvalues=(
        (
            subtitle.Subtitle(language='foo', forced=False, format=None),
            subtitle.Subtitle(language='FOO', forced=False, format=None),
            True,
        ),
        (
            subtitle.Subtitle(language='foo', forced=False, format=None),
            subtitle.Subtitle(language='bar', forced=False, format=None),
            False,
        ),
        (
            subtitle.Subtitle(language='foo', forced=False, format=None),
            subtitle.Subtitle(language='foo', forced=0, format=None),
            True,
        ),
        (
            subtitle.Subtitle(language='foo', forced=False, format=None),
            subtitle.Subtitle(language='foo', forced=1, format=None),
            False,
        ),
        (
            subtitle.Subtitle(language='foo', forced=False, format=None),
            subtitle.Subtitle(language='foo', forced=False, format=''),
            True,
        ),
        (
            subtitle.Subtitle(language='foo', forced=False, format='FMT'),
            subtitle.Subtitle(language='foo', forced=False, format='FMT2'),
            False,
        ),
    ),
)
def test_Subtitle___eq__(a, b, exp_equal):
    equal = a == b
    assert equal is exp_equal


@pytest.mark.parametrize(
    argnames='kwargs, exp_return_value',
    argvalues=(
        ({'language': 'en', 'forced': False, 'format': 'FMT'}, subtitle.Subtitle(language='en', forced=False, format='FMT')),
        ({'language': 'invalidcode', 'forced': False, 'format': 'FMT'}, None),
    ),
    ids=lambda v: repr(v),
)
def test_Subtitle_or_none(kwargs, exp_return_value):
    return_value = subtitle.Subtitle.or_none(**kwargs)
    assert return_value == exp_return_value


def test_get_subtitles(mocker):
    mocks = Mock()
    for funcname in dir(subtitle):
        if funcname.startswith('get_subtitles_from_'):
            mocks.attach_mock(
                mocker.patch(f'upsies.utils.subtitle.{funcname}', return_value={funcname, 'return', 'values'}),
                funcname,
            )

    content_path = 'path/to/content'
    subtitles = subtitle.get_subtitles(content_path)
    assert subtitles == {
        'get_subtitles_from_mediainfo',
        'get_subtitles_from_filenames',
        'get_subtitles_from_idx_files',
        'get_subtitles_from_dvd_tree',
        'return', 'values',
    }

    assert mocks.mock_calls == [
        call.get_subtitles_from_mediainfo(content_path),
        call.get_subtitles_from_filenames(content_path),
        call.get_subtitles_from_idx_files(content_path),
        call.get_subtitles_from_dvd_tree(content_path),
    ]


@pytest.mark.parametrize(
    argnames='all_tracks, exp_subtitles',
    argvalues=(
        ({'Video': ..., 'Audio': ...}, set()),
        ({'Video': ..., 'Audio': ..., 'Text': []}, set()),
        ({'Video': ..., 'Audio': ..., 'Text': [{}]}, set()),
        ({'Video': ..., 'Audio': ..., 'Text': [{'Language': 'invalidcode'}]}, set()),
        (
            {'Video': ..., 'Audio': ..., 'Text': [{'Language': 'foo'}]},
            {subtitle.Subtitle(language='foo', forced=False, format=None)},
        ),
        (
            {'Video': ..., 'Audio': ..., 'Text': [{'Language': 'foo', 'Forced': 'No', 'CodecID': 'unknown'}]},
            {subtitle.Subtitle(language='foo', forced=False, format='')},
        ),
        (
            {'Video': ..., 'Audio': ..., 'Text': [{'Language': 'foo', 'Forced': 'Yes', 'CodecID': 'S_TEXT/ASS'}]},
            {subtitle.Subtitle(language='foo', forced=True, format='ASS')},
        ),
        (
            {'Video': ..., 'Audio': ..., 'Text': [
                {'Language': 'foo', 'Forced': 'No', 'CodecID': 'S_HDMV/PGS'},
                {'Language': 'bar', 'Forced': 'No', 'CodecID': 'S_TEXT/ASS'},
                {'Language': 'baz', 'Forced': 'Yes', 'CodecID': 'S_TEXT/UTF8'},
            ]},
            {
                subtitle.Subtitle(language='foo', forced=False, format='PGS'),
                subtitle.Subtitle(language='bar', forced=False, format='ASS'),
                subtitle.Subtitle(language='baz', forced=True, format='SRT'),
            },
        ),
    ),
    ids=lambda v: repr(v),
)
def test_get_subtitles_from_mediainfo(all_tracks, exp_subtitles, mocker):
    tracks_mock = mocker.patch('upsies.utils.video.tracks', return_value=all_tracks)

    content_path = 'path/to/content'
    subtitles = subtitle.get_subtitles_from_mediainfo(content_path)
    assert subtitles == exp_subtitles

    assert tracks_mock.call_args_list == [call(content_path)]


@pytest.mark.parametrize(
    argnames='file_list, exp_subtitles',
    argvalues=(
        ([], set()),
        (['path/to/foo.mkv', 'path/to/foo.jpg'], set()),
        (['path/to/foo.mkv', 'path/to/foo.srt', 'path/to/foo.ssa', 'path/to/foo.ass', 'path/to/foo.vtt'], set()),
        (['path/to/foo.mkv', 'path/to/foo.en.srt', 'path/to/foo.ssa', 'path/to/foo.ass', 'path/to/foo.vtt'], {
            subtitle.Subtitle(language='en', forced=False, format='SRT'),
        }),
        (['path/to/foo.mkv', 'path/to/foo.en.srt', 'path/to/foo.fre.ssa', 'path/to/foo.ass', 'path/to/foo.vtt'], {
            subtitle.Subtitle(language='en', forced=False, format='SRT'),
            subtitle.Subtitle(language='fre', forced=False, format='SSA'),
        }),
        (['path/to/foo.mkv', 'path/to/foo.en.srt', 'path/to/foo.fre.ssa', 'path/to/foo.invalidcode.ass', 'path/to/foo.vtt'], {
            subtitle.Subtitle(language='en', forced=False, format='SRT'),
            subtitle.Subtitle(language='fre', forced=False, format='SSA'),
        }),
        (['path/to/foo.mkv', 'path/to/foo.en.srt', 'path/to/foo.fre.ssa', 'path/to/foo.pt.ass', 'path/to/foo.vie.vtt'], {
            subtitle.Subtitle(language='en', forced=False, format='SRT'),
            subtitle.Subtitle(language='fre', forced=False, format='SSA'),
            subtitle.Subtitle(language='pt', forced=False, format='ASS'),
            subtitle.Subtitle(language='vie', forced=False, format='VTT'),
        }),
    ),
    ids=lambda v: repr(v),
)
def test_get_subtitles_from_filenames(file_list, exp_subtitles, mocker):
    file_list_mock = mocker.patch('upsies.utils.fs.file_list', return_value=file_list)

    content_path = 'path/to/content'
    subtitles = subtitle.get_subtitles_from_filenames(content_path)
    assert subtitles == exp_subtitles

    assert file_list_mock.call_args_list == [
        call(content_path, extensions=('srt', 'ssa', 'ass', 'vtt')),
    ]


@pytest.mark.parametrize(
    argnames='existing_files, file_list, exp_subtitles',
    argvalues=(
        ({}, [], set()),
        ({'foo.idx': 'junk\nmore junk\n'}, ['foo.idx'], set()),
        ({'foo.idx': 'junk\nid: invalidcode, index: 0\nmore junk\n'}, ['foo.idx'], set()),
        (
            {
                'foo.idx': 'junk\nid: de, index: 0\nmore junk\n',
                'bar.idx': 'junk\nid: en, index: 1\nmore junk\n',
                'baz.idx': 'junk\nid: ru, index: 123\nmore junk\n',
            },
            ['foo.idx', 'bar.idx', 'baz.idx', 'nosuchfile.idx'],
            {
                subtitle.Subtitle(language='de', forced=False, format=subtitle._subtitle_formats['idx']),
                subtitle.Subtitle(language='en', forced=False, format=subtitle._subtitle_formats['idx']),
                subtitle.Subtitle(language='ru', forced=False, format=subtitle._subtitle_formats['idx']),
            },
        ),
    ),
    ids=lambda v: repr(v),
)
def test_get_subtitles_from_idx_files(existing_files, file_list, exp_subtitles, tmp_path, mocker):
    content_path = tmp_path / 'content'
    content_path.mkdir(parents=True, exist_ok=True)
    for file, data in existing_files.items():
        (content_path / file).write_text(data)

    file_list_mock = mocker.patch('upsies.utils.fs.file_list', return_value=(
        str(content_path / file)
        for file in file_list
    ))

    subtitles = subtitle.get_subtitles_from_idx_files(content_path)
    assert subtitles == exp_subtitles

    assert file_list_mock.call_args_list == [
        call(content_path, extensions=('idx',)),
    ]


@pytest.mark.parametrize(
    argnames='found_videos, exp_subtitles',
    argvalues=(
        (
            {
                'a.mkv': {'Video': ..., 'Audio': ..., 'Text': [{'Language': 'a', 'Forced': 'Yes', 'CodecID': 'S_TEXT/UTF8'}]},
                'VTS_01_02.VOB': {'Video': ..., 'Audio': ..., 'Text': [
                    {'Language': 'xa'},
                    {'Language': 'xb'},
                ]},
                'VIDEO_TS.IFO': {},
                'b.jpg': {},
                'VTS_03_04.VOB': {'Video': ..., 'Audio': ..., 'Text': [
                    {'Language': 'ya'},
                    {'Language': 'yb'},
                ]},
            },
            set((
                subtitle.Subtitle(language='xa', forced=False, format='VobSub'),
                subtitle.Subtitle(language='xb', forced=False, format='VobSub'),
                subtitle.Subtitle(language='ya', forced=False, format='VobSub'),
                subtitle.Subtitle(language='yb', forced=False, format='VobSub'),
            )),
        ),
    ),
    ids=lambda v: repr(v),
)
def test_get_subtitles_from_dvd_tree(found_videos, exp_subtitles, tmp_path, mocker):
    content_path = 'path/to/content'

    find_videos_mock = mocker.patch('upsies.utils.video.find_videos', return_value=tuple(found_videos))
    tracks_mock = mocker.patch('upsies.utils.video.tracks', side_effect=(
        tracks
        for filename, tracks in found_videos.items()
        if filename.upper().endswith('VOB')
    ))

    subtitles = subtitle.get_subtitles_from_dvd_tree(content_path)
    assert subtitles == exp_subtitles

    assert find_videos_mock.call_args_list == [
        call(content_path),
    ]
    assert tracks_mock.call_args_list == [
        call(re.sub(r'\d+.VOB', '0.IFO', filename))
        for filename in found_videos.keys()
        if filename.upper().endswith('VOB')
    ]



"""
Print ``mediainfo`` output
"""

import functools

from .... import jobs, utils
from .base import CommandBase


class mediainfo(CommandBase):
    """
    Print ``mediainfo`` output

    Directories are recursively searched for the first video file in natural
    order, e.g. "File1.mp4" comes before "File10.mp4".

    Any irrelevant leading parts in the file path are removed from the output.
    """

    names = ('mediainfo', 'mi')

    argument_definitions = {
        'CONTENT': {
            'type': utils.argtypes.content,
            'help': 'Path to release content',
        },
        ('--from-all-videos', '-a'): {
            'action': 'store_true',
            'help': 'Get mediainfo from each video file beneath CONTENT',
        },
        ('--exclude-files', '--ef'): {
            'nargs': '+',
            'metavar': 'PATTERN',
            'help': ('Glob pattern to exclude from CONTENT '
                     '(matched case-insensitively against path relative to CONTENT)'),
            'default': (),
        },
        ('--exclude-files-regex', '--efr'): {
            'nargs': '+',
            'metavar': 'PATTERN',
            'help': ('Regular expression to exclude from CONTENT '
                     '(matched case-sensitively against path relative to CONTENT)'),
            'type': utils.argtypes.regex,
            'default': (),
        },
    }

    @functools.cached_property
    def jobs(self):
        return (
            jobs.mediainfo.MediainfoJob(
                home_directory=self.home_directory,
                cache_directory=self.cache_directory,
                ignore_cache=self.args.ignore_cache,
                content_path=self.args.CONTENT,
                from_all_videos=self.args.from_all_videos,
                exclude_files=(
                    tuple(self.args.exclude_files)
                    + tuple(self.args.exclude_files_regex)
                ),
            ),
        )
